# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import itertools
import os
import shutil

import pytest
import yaml

from swh.model.model import Content
from swh.model.swhids import ExtendedSWHID

from ..recovery_bundle import (
    AgeSecretKey,
    Manifest,
    RecoveryBundle,
    RecoveryBundleCreator,
    SecretRecoveryError,
    SecretSharing,
    age_decrypt,
    age_encrypt,
    convert_bits,
    generate_age_keypair,
    list_yubikey_identities,
    recover_object_decryption_key_from_encrypted_shares,
)


@pytest.fixture
def manifest_dict():
    return {
        "version": 1,
        "removal_identifier": "TDN-2023-06-18-01",
        "created": datetime.datetime(
            2023, 6, 18, 13, 12, 42, tzinfo=datetime.timezone.utc
        ),
        "swhids": [
            "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
            "swh:1:snp:0000000000000000000000000000000000000022",
            "swh:1:rel:0000000000000000000000000000000000000021",
            "swh:1:rev:0000000000000000000000000000000000000018",
            "swh:1:rev:0000000000000000000000000000000000000013",
            "swh:1:dir:0000000000000000000000000000000000000017",
            "swh:1:cnt:0000000000000000000000000000000000000016",
            "swh:1:cnt:0000000000000000000000000000000000000012",
            "swh:1:cnt:0000000000000000000000000000000000000015",
            "swh:1:cnt:0000000000000000000000000000000000000014",
            "swh:1:cnt:0000000000000000000000000000000000000011",
        ],
        "decryption_key_shares": {
            "YubiKey serial 4245067 slot 1": "-----BEGIN AGE ENCRYPTED FILE-----\n"
            "YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IHBpdi1wMjU2IHcvb0k0USBBb3FMYjRM\n"
            "V3dlcm9YazZkTU9UZld4eEVhYUlBZHRBQ05CQndOUFZJMmV1NApmNTY1MUJFdks1\n"
            "aE9TZzQ3NFJGN0cvQlFIMDZNSTkxUEpOblJteUkyK2FVCi0+IDxYTSFKLWdyZWFz\n"
            "ZSBCfWErZHkKNEMrbTdqekhTZTQ4c3pXRGZjK3N0UTh2Qi9ISU1XdFF6a0RvdmRl\n"
            "NAotLS0gYk9Ob2dkUTJRZE9nT3BTK29JWU5pRkZIVC9pUzJQaHRZc05sMjd6S1Rr\n"
            "OAoRXkzBiNX98H+353sOjGxJvCdYmtUdn7ozR35g+VSB6zxS972s2drkuKxQ0kIN\n"
            "MIjaytf/RJ0J3N/x8CtsEvXSoGjnuIT0GuEUbCqG0Qg0/YrrDzEGcD34l6JnD187\n"
            "5nVFnUimLXK6S2HeEDTJUZuLWfmglqaZaZjPnEKxqu8TfrJDBgg7miJLC+rGXhn9\n"
            "4ArtFIaOQgotCHZ8Y0lpmqGJIVTKWgdgpW+JjzyG\n"
            "-----END AGE ENCRYPTED FILE-----\n",
            "YubiKey serial 5229836 slot 1": "-----BEGIN AGE ENCRYPTED FILE-----\n"
            "YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IHBpdi1wMjU2IHcvb0k0USBBaTZhaUo3\n"
            "WnMzMmlTUlp5QmNhTkI1bHlmcHNyY0FPQ0RnK1BQdHQxS0EvbAppVnExb3BZcFRW\n"
            "ZkZ1ZFZrQWlyaU9HTkRKREYvU2tSaldkSHpWdVd1aGFVCi0+IDUrPVssLWdyZWFz\n"
            "ZQpzcm1WSkNqOWVrOU5GUXRMSmpFVVR4aEhrM0UKLS0tIFl2QkN6d1QzdWN6U0dB\n"
            "VHVzYk1SdDBLNlhNanJGc2x4L2hMZTZrSUxTSGMKLOKIpGZtKtUeOsSrcoIvKiBu\n"
            "DAoLXMGY+302lQRJsdJ3I7N+eFhRATsOM7vO8eupXbee87kIkGB7GaqGR5X48GR1\n"
            "oNrMsY5PcjZICxLjWYX9cMVMAXcmBjV9ZCWwqzmw86rY0k74mRwhE0dYd95P90+5\n"
            "NniuNgxQYKkM5QoKVHn36ISJGUgcvp5/JCM69X7kM8UvjLarFeYdHfqqAZUImNla\n"
            "lEdIqdOmnUs=\n"
            "-----END AGE ENCRYPTED FILE-----\n",
        },
        "reason": "copyright issue",
        "expire": datetime.datetime(
            2024, 6, 18, 13, 12, 42, tzinfo=datetime.timezone.utc
        ),
    }


def test_manifest_load_success(manifest_dict):
    assert Manifest.load(yaml.dump(manifest_dict))


def test_manifest_load_success_with_no_optionals(manifest_dict):
    del manifest_dict["reason"]
    del manifest_dict["expire"]
    assert Manifest.load(yaml.dump(manifest_dict))


@pytest.mark.parametrize(
    "invalid_manifest_dict",
    [
        pytest.param({"version": 2}, id="invalid_version"),
        pytest.param({"expire": "2024-06-18T13:12:42Z"}, id="str_instead_of_datetime"),
        pytest.param({"swhids": []}, id="empty_swhids"),
        pytest.param({"decryption_key_shares": {}}, id="empty_shares"),
        pytest.param({"invalid": "field"}, id="invalid_field"),
    ],
)
def test_manifest_load_failure(manifest_dict, invalid_manifest_dict):
    manifest_dict.update(invalid_manifest_dict)
    with pytest.raises((ValueError, TypeError)):
        assert Manifest.load(yaml.dump(manifest_dict))


EXPECTED_MANIFEST_DUMP = """\
version: 1
removal_identifier: TDN-2023-06-18-01
created: 2023-06-18T13:12:42+00:00
swhids:
- swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165
- swh:1:snp:0000000000000000000000000000000000000022
- swh:1:rel:0000000000000000000000000000000000000021
- swh:1:rev:0000000000000000000000000000000000000018
- swh:1:rev:0000000000000000000000000000000000000013
- swh:1:dir:0000000000000000000000000000000000000017
- swh:1:cnt:0000000000000000000000000000000000000016
- swh:1:cnt:0000000000000000000000000000000000000012
- swh:1:cnt:0000000000000000000000000000000000000015
- swh:1:cnt:0000000000000000000000000000000000000014
- swh:1:cnt:0000000000000000000000000000000000000011
decryption_key_shares:
  YubiKey serial 4245067 slot 1: |
    -----BEGIN AGE ENCRYPTED FILE-----
    YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IHBpdi1wMjU2IHcvb0k0USBBb3FMYjRM
    V3dlcm9YazZkTU9UZld4eEVhYUlBZHRBQ05CQndOUFZJMmV1NApmNTY1MUJFdks1
    aE9TZzQ3NFJGN0cvQlFIMDZNSTkxUEpOblJteUkyK2FVCi0+IDxYTSFKLWdyZWFz
    ZSBCfWErZHkKNEMrbTdqekhTZTQ4c3pXRGZjK3N0UTh2Qi9ISU1XdFF6a0RvdmRl
    NAotLS0gYk9Ob2dkUTJRZE9nT3BTK29JWU5pRkZIVC9pUzJQaHRZc05sMjd6S1Rr
    OAoRXkzBiNX98H+353sOjGxJvCdYmtUdn7ozR35g+VSB6zxS972s2drkuKxQ0kIN
    MIjaytf/RJ0J3N/x8CtsEvXSoGjnuIT0GuEUbCqG0Qg0/YrrDzEGcD34l6JnD187
    5nVFnUimLXK6S2HeEDTJUZuLWfmglqaZaZjPnEKxqu8TfrJDBgg7miJLC+rGXhn9
    4ArtFIaOQgotCHZ8Y0lpmqGJIVTKWgdgpW+JjzyG
    -----END AGE ENCRYPTED FILE-----
  YubiKey serial 5229836 slot 1: |
    -----BEGIN AGE ENCRYPTED FILE-----
    YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IHBpdi1wMjU2IHcvb0k0USBBaTZhaUo3
    WnMzMmlTUlp5QmNhTkI1bHlmcHNyY0FPQ0RnK1BQdHQxS0EvbAppVnExb3BZcFRW
    ZkZ1ZFZrQWlyaU9HTkRKREYvU2tSaldkSHpWdVd1aGFVCi0+IDUrPVssLWdyZWFz
    ZQpzcm1WSkNqOWVrOU5GUXRMSmpFVVR4aEhrM0UKLS0tIFl2QkN6d1QzdWN6U0dB
    VHVzYk1SdDBLNlhNanJGc2x4L2hMZTZrSUxTSGMKLOKIpGZtKtUeOsSrcoIvKiBu
    DAoLXMGY+302lQRJsdJ3I7N+eFhRATsOM7vO8eupXbee87kIkGB7GaqGR5X48GR1
    oNrMsY5PcjZICxLjWYX9cMVMAXcmBjV9ZCWwqzmw86rY0k74mRwhE0dYd95P90+5
    NniuNgxQYKkM5QoKVHn36ISJGUgcvp5/JCM69X7kM8UvjLarFeYdHfqqAZUImNla
    lEdIqdOmnUs=
    -----END AGE ENCRYPTED FILE-----
reason: copyright issue
expire: 2024-06-18T13:12:42+00:00
"""


def test_manifest_dump(manifest_dict):
    manifest = Manifest(**manifest_dict)
    assert manifest.dump() == EXPECTED_MANIFEST_DUMP


def test_manifest_dump_no_optionals(manifest_dict):
    del manifest_dict["reason"]
    del manifest_dict["expire"]
    manifest = Manifest(**manifest_dict)
    dump = manifest.dump()
    assert "reason:" not in dump
    assert "expire:" not in dump


def test_manifest_load():
    manifest = Manifest.load(EXPECTED_MANIFEST_DUMP)
    assert isinstance(manifest, Manifest)


def test_manifest_dump_load_roundtrip(manifest_dict):
    manifest = Manifest(**manifest_dict)
    assert manifest == Manifest.load(manifest.dump())


def test_generate_age_keypair():
    public_key, secret_key = generate_age_keypair()
    assert public_key.startswith("age")
    assert len(public_key) == 62
    assert secret_key.startswith("AGE-SECRET-KEY-")
    assert len(secret_key) == 74


OBJECT_PUBLIC_KEY = "age1a4uwpku4xzlnkh78ma3urlulhhhz0xlsv6crthjvhrjysvskp9nsz77qts"
OBJECT_SECRET_KEY = (
    "AGE-SECRET-KEY-1EZMJLS2MMEN4D6CCR6TQ66RD4MPT32ZN8EAU44PS3EDNUAKQWE0SM92NN4"
)


def test_age_encrypt_decrypt_roundtrip():
    public_key = OBJECT_PUBLIC_KEY
    secret_key = OBJECT_SECRET_KEY
    cleartext = b"All Your Source Code Base Must Belong To Us!"
    ciphertext = age_encrypt(public_key, cleartext)
    assert age_decrypt(secret_key, ciphertext) == cleartext


def test_convert_5bits_to_8bits():
    assert [104, 101, 108, 108, 111] == convert_bits(
        [13, 1, 18, 22, 24, 27, 3, 15], 5, 8
    )


def test_convert_8bits_to_5bits():
    assert convert_bits([104, 101, 108, 108, 111], 8, 5) == [
        13,
        1,
        18,
        22,
        24,
        27,
        3,
        15,
    ]


EXAMPLE_SECRET_SHARING_YAML = """\
secret_sharing:
  minimum_required_groups: 2
  groups:
    legal:
      minimum_required_shares: 1
      recipient_keys:
        "YubiKey serial 4245067 slot 1": |-
          age1yubikey1q2e37f74zzazz75mtggzql3at66pegemfnul0dtd7axctahljkvsqezscaq
        "YubiKey serial 2284622 slot 3": |-
          age1yubikey1q0ucnwg558zcwrc752evk3620q2t4mkwz6a0lq9u3clsfmealsmlz330kz2
    sysadmins:
      minimum_required_shares: 1
      recipient_keys:
        "YubiKey serial 3862152 slot 1": |-
          age1yubikey1q04k5fs8cz6kypt7vjetl2gc2qtpz9lyzpxrvv2agzt6h8n3awmzk9sgd8v
        "YubiKey serial 6927448 slot 1": |-
          age1yubikey1qt2p377vq6qg58l8gaframp9yggvsysddraa72aehma5mw623r8rqk0mlgu
"""


@pytest.fixture
def example_secret_sharing():
    return SecretSharing.from_dict(
        yaml.safe_load(EXAMPLE_SECRET_SHARING_YAML)["secret_sharing"]
    )


@pytest.mark.skipif(
    not shutil.which("age-plugin-yubikey"),
    reason="missing `age-plugin-yubikey` executable",
)
def test_generate_encrypted_shares(example_secret_sharing):
    encrypted_shares = example_secret_sharing.generate_encrypted_shares(
        "example", OBJECT_SECRET_KEY
    )
    assert len(encrypted_shares) == 4
    assert encrypted_shares.keys() == {
        "YubiKey serial 4245067 slot 1",
        "YubiKey serial 2284622 slot 3",
        "YubiKey serial 3862152 slot 1",
        "YubiKey serial 6927448 slot 1",
    }
    assert all(
        s.startswith("-----BEGIN AGE ENCRYPTED FILE-----")
        for s in encrypted_shares.values()
    )


ALI_PUBLIC_KEY = "age123hpq9m25xsmx7caqvyv8k3fxaqastc3evyq9q7myur7l9ukj4dsnp7a5v"
ALI_SECRET_KEY = (
    "AGE-SECRET-KEY-1VREXCYE5WNMUD0WSCF7F6CH3FGQ9P6PGD25QHY7QX8PGDN87P37QQD3L2G"
)
BOB_PUBLIC_KEY = "age1mrhte5tlpzpz57gg85nzcefqc5pm5usmakqpuurxux7ry2rmhdgs7r9u68"
BOB_SECRET_KEY = (
    "AGE-SECRET-KEY-1UPJU3AF4M0NPLSLGVDWJU38F3MDE3JJM48E8NST8V3YKU077HEVSQVPZC2"
)
CAMILLE_PUBLIC_KEY = "age1ahuqxgjmvfm65shmwqa7xa703vvcla528swga3zempnxslj3pczqtx6wr8"
CAMILLE_SECRET_KEY = (
    "AGE-SECRET-KEY-1NPLST9VXL6E9DEHCVVPUTGH60ZRJFLPZ5HDM93MJW993CGFQ49PQU90RVL"
)
DLIQUE_PUBLIC_KEY = "age1qwu50kncctmpky7gg5s0v4mt4fzc4wjwj6mfjzhtk3wn6pspkyksmsmhze"
DLIQUE_SECRET_KEY = (
    "AGE-SECRET-KEY-1NPT3PFA7N03DFQY9GN764T4TJCZSLP36YV4S98FLYN0YGX2539GSSFUT4F"
)
ESSUN_PUBLIC_KEY = "age1uakt638m65nt56q9qjecwp60gnv6qwqkez43re06awqzf8hqh3pqnsppaw"
ESSUN_SECRET_KEY = (
    "AGE-SECRET-KEY-10ZZWX7FNCUJR7HRACEGUCVA4V0PYGLQ7NJDYPRH96YNC3AJLM37QQNWX3K"
)

TWO_GROUPS_REQUIRED_WITH_ONE_MINIMUM_SHARE_EACH_SECRET_SHARING_YAML = f"""\
secret_sharing:
  minimum_required_groups: 2
  groups:
    legal:
      minimum_required_shares: 1
      recipient_keys:
        "Ali": {ALI_PUBLIC_KEY}
        "Bob": {BOB_PUBLIC_KEY}
    sysadmins:
      minimum_required_shares: 1
      recipient_keys:
        "Camille": {CAMILLE_PUBLIC_KEY}
        "Dlique": {DLIQUE_PUBLIC_KEY}
"""


@pytest.fixture
def secret_sharing_2_groups_required_with_1_minimum_each():
    return SecretSharing.from_dict(
        yaml.safe_load(
            TWO_GROUPS_REQUIRED_WITH_ONE_MINIMUM_SHARE_EACH_SECRET_SHARING_YAML
        )["secret_sharing"]
    )


def available_secret_keys_for_2_groups_required_with_1_minimum_each():
    yield ("Dlique", DLIQUE_SECRET_KEY)
    yield ("Ali", ALI_SECRET_KEY)


TWO_GROUPS_REQUIRED_OF_THREE_WITH_ONE_AND_TWO_MINIMUM_IN_EACH_SECRET_SHARING_YAML = f"""\
secret_sharing:
  minimum_required_groups: 2
  groups:
    legal:
      minimum_required_shares: 1
      recipient_keys:
        "Ali": {ALI_PUBLIC_KEY}
    dpo:
      minimum_required_shares: 1
      recipient_keys:
        "Bob": {BOB_PUBLIC_KEY}
    sysadmins:
      minimum_required_shares: 2
      recipient_keys:
        "Camille": {CAMILLE_PUBLIC_KEY}
        "Dlique": {DLIQUE_PUBLIC_KEY}
        "Essun": {ESSUN_PUBLIC_KEY}
"""


@pytest.fixture
def secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each():
    return SecretSharing.from_dict(
        yaml.safe_load(
            TWO_GROUPS_REQUIRED_OF_THREE_WITH_ONE_AND_TWO_MINIMUM_IN_EACH_SECRET_SHARING_YAML
        )["secret_sharing"]
    )


def available_secret_keys_for_2_groups_required_of_3_only_ones():
    yield ("Ali", ALI_SECRET_KEY)
    yield ("Bob", BOB_SECRET_KEY)


def available_secret_keys_for_2_groups_required_of_3_one_and_two():
    yield ("Essun", ESSUN_SECRET_KEY)
    yield ("Bob", BOB_SECRET_KEY)
    yield ("Camille", CAMILLE_SECRET_KEY)


@pytest.mark.parametrize(
    "secret_sharing, available_secret_keys",
    [
        (
            "secret_sharing_2_groups_required_with_1_minimum_each",
            available_secret_keys_for_2_groups_required_with_1_minimum_each,
        ),
        (
            "secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each",
            available_secret_keys_for_2_groups_required_of_3_only_ones,
        ),
        (
            "secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each",
            available_secret_keys_for_2_groups_required_of_3_one_and_two,
        ),
    ],
)
def test_object_decryption_key_recovery_roundtrip(
    request, secret_sharing, available_secret_keys
):
    secret_sharing = request.getfixturevalue(secret_sharing)

    encrypted_shares = secret_sharing.generate_encrypted_shares(
        secret_sharing, OBJECT_SECRET_KEY
    )
    recovered_key = recover_object_decryption_key_from_encrypted_shares(
        encrypted_shares, available_secret_keys
    )
    assert recovered_key == OBJECT_SECRET_KEY


def test_object_decryption_key_recovery_with_not_enough_secret_keys(
    secret_sharing_2_groups_required_with_1_minimum_each,
):
    def available_secret_keys():
        yield ("Dlique", DLIQUE_SECRET_KEY)

    secret_sharing = secret_sharing_2_groups_required_with_1_minimum_each
    encrypted_shares = secret_sharing.generate_encrypted_shares(
        "2-groups-required-with-1-minimum-each", OBJECT_SECRET_KEY
    )
    with pytest.raises(SecretRecoveryError):
        _ = recover_object_decryption_key_from_encrypted_shares(
            encrypted_shares, available_secret_keys
        )


def test_object_decryption_key_recovery_with_known_shares(
    secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each,
):
    def available_secret_keys():
        yield ("Essun", ESSUN_SECRET_KEY)

    secret_sharing = (
        secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each
    )
    encrypted_shares = secret_sharing.generate_encrypted_shares(
        "2-groups-required-of-3-with-1-and-two-minimum-in-each", OBJECT_SECRET_KEY
    )
    camille_mnemonic = age_decrypt(
        CAMILLE_SECRET_KEY, encrypted_shares["Camille"]
    ).decode("us-ascii")
    bob_mnemonic = age_decrypt(BOB_SECRET_KEY, encrypted_shares["Bob"]).decode(
        "us-ascii"
    )
    recovered_key = recover_object_decryption_key_from_encrypted_shares(
        encrypted_shares,
        available_secret_keys,
        known_mnemonics=[
            camille_mnemonic,
            bob_mnemonic,
        ],
    )
    assert recovered_key == OBJECT_SECRET_KEY


def test_secret_sharing_errors_with_duplicate_identifiers():
    conf = f"""\
        minimum_required_groups: 2
        groups:
          legal:
            minimum_required_shares: 1
            recipient_keys:
              "Ali": {ALI_PUBLIC_KEY}
              "Bob": {BOB_PUBLIC_KEY}
          sysadmins:
            minimum_required_shares: 1
            recipient_keys:
              # Identifier also present in the legal group
              "Ali": {CAMILLE_PUBLIC_KEY}
              "Dlique": {DLIQUE_PUBLIC_KEY}
    """
    with pytest.raises(ValueError, match="Duplicate share identifier"):
        _ = SecretSharing.from_dict(yaml.safe_load(conf))


def test_secret_sharing_errors_with_duplicate_keys():
    conf = f"""\
        minimum_required_groups: 2
        groups:
          legal:
            minimum_required_shares: 1
            recipient_keys:
              "Ali": {ALI_PUBLIC_KEY}
              "Bob": {BOB_PUBLIC_KEY}
          sysadmins:
            minimum_required_shares: 1
            recipient_keys:
              # Recipient key is the same as Ali’s
              "Camille": {ALI_PUBLIC_KEY}
              "Dlique": {DLIQUE_PUBLIC_KEY}
    """
    with pytest.raises(ValueError, match="Duplicate recipient public key"):
        _ = SecretSharing.from_dict(yaml.safe_load(conf))


MOCK_AGE_PLUGIN_YUBIKEY_SCRIPT = """\
#!/bin/sh

if [ "$1" != "--identity" ]; then
    exit 1
fi

cat <<EOF
#       Serial: 5229836, Slot: 1
#         Name: age identity c3fa08e1
#      Created: Wed, 28 Jun 2023 12:24:19 +0000
#   PIN policy: Never  (A PIN is NOT required to decrypt)
# Touch policy: Cached (A physical touch is required for decryption, and is cached for 15 seconds)
#    Recipient: age1yubikey1q2e37f74zzazz75mtggzql3at66pegemfnul0dtd7axctahljkvsqezscaq
AGE-PLUGIN-YUBIKEY-1PNX57QYZC0AQ3CGZL7UL3


#       Serial: 4245067, Slot: 1
#         Name: age identity 4c1bc5f1
#      Created: Wed, 28 Jun 2023 14:43:01 +0000
#   PIN policy: Never  (A PIN is NOT required to decrypt)
# Touch policy: Always (A physical touch is required for every decryption)
#    Recipient: age1yubikey1q04k5fs8cz6kypt7vjetl2gc2qtpz9lyzpxrvv2agzt6h8n3awmzk9sgd8v
AGE-PLUGIN-YUBIKEY-1F0RYQQYZFSDUTUG4RNAMD
EOF
"""  # noqa: B950


@pytest.fixture
def mock_age_yubikey_plugin(tmp_path):
    script_path = tmp_path / "age-plugin-yubikey"
    script_path.write_text(MOCK_AGE_PLUGIN_YUBIKEY_SCRIPT)
    os.chmod(script_path, 0o500)
    old_path = os.environ["PATH"]
    os.environ["PATH"] = f"{tmp_path}:{os.environ['PATH']}"
    yield
    os.environ["PATH"] = old_path


def test_list_yubikey_identities(mock_age_yubikey_plugin):
    assert list_yubikey_identities() == [
        ("YubiKey serial 5229836 slot 1", "AGE-PLUGIN-YUBIKEY-1PNX57QYZC0AQ3CGZL7UL3"),
        ("YubiKey serial 4245067 slot 1", "AGE-PLUGIN-YUBIKEY-1F0RYQQYZFSDUTUG4RNAMD"),
    ]


@pytest.fixture
def encrypted_shares_for_object_private_key(
    secret_sharing_2_groups_required_with_1_minimum_each,
):
    secret_sharing = secret_sharing_2_groups_required_with_1_minimum_each
    return secret_sharing.generate_encrypted_shares(
        "2-groups-with-1-minimum-each", OBJECT_SECRET_KEY
    )


@pytest.fixture
def sample_populated_storage(swh_storage, sample_data):
    swh_storage.content_add(sample_data.contents)
    swh_storage.skipped_content_add(sample_data.skipped_contents)
    swh_storage.directory_add(sample_data.directories)
    swh_storage.revision_add(sample_data.git_revisions)
    swh_storage.release_add(sample_data.releases)
    swh_storage.snapshot_add(sample_data.snapshots)
    swh_storage.origin_add(sample_data.origins)
    swh_storage.origin_visit_add(sample_data.origin_visits)
    swh_storage.origin_visit_status_add(sample_data.origin_visit_statuses)
    return swh_storage


def test_create_recovery_bundle(
    tmp_path, sample_populated_storage, encrypted_shares_for_object_private_key
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    swhids = [
        # Content
        "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
        "swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea",
        # SkippedContent
        "swh:1:cnt:33e45d56f88993aae6a0198013efa80716fd8920",
        # Directory
        "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
        "swh:1:dir:4b825dc642cb6eb9a060e54bf8d69288fbee4904",
        "swh:1:dir:afa0105cfcaa14fdbacee344e96659170bb1bda5",
        # Revision
        "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
        "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
        # Release
        "swh:1:rel:f7f222093a18ec60d781070abec4a630c850b837",
        "swh:1:rel:db81a26783a3f4a9db07b4759ffc37621f159bb2",
        # Snapshot
        "swh:1:snp:9b922e6d8d5b803c1582aabe5525b7b91150788e",
        "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
        # Origin
        "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645",
        "swh:1:ori:9147ab9c9287940d4fdbe95d8780664d7ad2dfc0",
    ]
    with RecoveryBundleCreator(
        path=bundle_path,
        storage=sample_populated_storage,
        removal_identifier="test_bundle",
        object_public_key=OBJECT_PUBLIC_KEY,
        decryption_key_shares=encrypted_shares_for_object_private_key,
    ) as creator:
        creator.backup_swhids(ExtendedSWHID.from_string(swhid) for swhid in swhids)

    from zipfile import ZipFile

    from swh.journal.serializers import kafka_to_value

    with ZipFile(bundle_path, "r") as bundle:
        # Do we have the expected files?
        assert bundle.namelist() == [
            "contents/swh_1_cnt_d81cc0710eb6cf9efd5b920a8453e1e07157b6cd.age",
            "contents/swh_1_cnt_c932c7649c6dfa4b82327d121215116909eb3bea.age",
            "skipped_contents/swh_1_cnt_33e45d56f88993aae6a0198013efa80716fd8920_1.age",
            "directories/swh_1_dir_5256e856a0a0898966d6ba14feb4388b8b82d302.age",
            "directories/swh_1_dir_4b825dc642cb6eb9a060e54bf8d69288fbee4904.age",
            "directories/swh_1_dir_afa0105cfcaa14fdbacee344e96659170bb1bda5.age",
            "revisions/swh_1_rev_01a7114f36fddd5ef2511b2cadda237a68adbb12.age",
            "revisions/swh_1_rev_a646dd94c912829659b22a1e7e143d2fa5ebde1b.age",
            "releases/swh_1_rel_f7f222093a18ec60d781070abec4a630c850b837.age",
            "releases/swh_1_rel_db81a26783a3f4a9db07b4759ffc37621f159bb2.age",
            "snapshots/swh_1_snp_9b922e6d8d5b803c1582aabe5525b7b91150788e.age",
            "snapshots/swh_1_snp_db99fda25b43dc5cd90625ee4b0744751799c917.age",
            "origins/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645.age",
            "origin_visits/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645_1.age",
            "origin_visit_statuses/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645_1_2015-01-01T23_00_00.000000+00_00.age",
            "origin_visits/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645_2.age",
            "origin_visit_statuses/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645_2_2017-01-01T23_00_00.000000+00_00.age",
            "origins/swh_1_ori_9147ab9c9287940d4fdbe95d8780664d7ad2dfc0.age",
            "origin_visits/swh_1_ori_9147ab9c9287940d4fdbe95d8780664d7ad2dfc0_1.age",
            "origin_visit_statuses/swh_1_ori_9147ab9c9287940d4fdbe95d8780664d7ad2dfc0_1_2015-01-01T23_00_00.000000+00_00.age",
            "manifest.yml",
        ]
        # Can we load the manifest?
        manifest = Manifest.load(bundle.read("manifest.yml"))
        assert isinstance(manifest, Manifest)
        assert manifest.removal_identifier == "test_bundle"
        # Can we unpack, decrypt and load at least an object?
        encrypted_serialized_content = bundle.read(
            "contents/swh_1_cnt_d81cc0710eb6cf9efd5b920a8453e1e07157b6cd.age"
        )
        serialized_content = age_decrypt(
            OBJECT_SECRET_KEY, encrypted_serialized_content
        )
        content: Content = Content.from_dict(kafka_to_value(serialized_content))
        assert content.status == "visible"
        # Have we properly saved content data?
        assert content.data == b"42\n"


def test_create_recovery_bundle_fails_if_empty(
    tmp_path, sample_populated_storage, encrypted_shares_for_object_private_key
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    with pytest.raises(ValueError):
        with RecoveryBundleCreator(
            path=bundle_path,
            storage=sample_populated_storage,
            removal_identifier="test_bundle",
            object_public_key=OBJECT_PUBLIC_KEY,
            decryption_key_shares=encrypted_shares_for_object_private_key,
        ) as _:
            # oops, we do not add any objects
            pass


def test_create_recovery_bundle_fails_without_decryption_key_shares(
    tmp_path, sample_populated_storage
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    swhids = ["swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645"]
    with pytest.raises(ValueError):
        with RecoveryBundleCreator(
            path=bundle_path,
            storage=sample_populated_storage,
            removal_identifier="test_bundle",
            object_public_key=OBJECT_PUBLIC_KEY,
            decryption_key_shares={},
        ) as creator:
            creator.backup_swhids(ExtendedSWHID.from_string(swhid) for swhid in swhids)


def test_create_recovery_bundle_with_optional_fields(
    tmp_path, sample_populated_storage, encrypted_shares_for_object_private_key
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    swhids = ["swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645"]
    expiration_date = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(
        days=365
    )
    with RecoveryBundleCreator(
        path=bundle_path,
        storage=sample_populated_storage,
        removal_identifier="test_bundle",
        object_public_key=OBJECT_PUBLIC_KEY,
        decryption_key_shares=encrypted_shares_for_object_private_key,
    ) as creator:
        creator.backup_swhids(ExtendedSWHID.from_string(swhid) for swhid in swhids)
        creator.set_reason("we are running a test")
        creator.set_expire(expiration_date)

    from zipfile import ZipFile

    with ZipFile(bundle_path, "r") as bundle:
        manifest = Manifest.load(bundle.read("manifest.yml"))
        assert manifest.reason == "we are running a test"
        assert manifest.expire.isoformat(
            timespec="seconds"
        ) == expiration_date.isoformat(timespec="seconds")


@pytest.fixture
def sample_recovery_bundle_path():
    return os.path.join(
        os.path.dirname(__file__), "fixtures", "sample.swh-recovery-bundle"
    )


def object_decryption_key_provider_for_sample(manifest: Manifest) -> AgeSecretKey:
    return OBJECT_SECRET_KEY


@pytest.fixture
def sample_recovery_bundle(sample_recovery_bundle_path):
    return RecoveryBundle(
        sample_recovery_bundle_path, object_decryption_key_provider_for_sample
    )


def test_recovery_bundle_decryption_key_provider_is_optional(
    sample_recovery_bundle_path,
):
    bundle = RecoveryBundle(sample_recovery_bundle_path)
    # We can access the manifest even without a decryption_key_provider
    assert bundle.removal_identifier == "test_bundle"
    # But trying to get the object_decryption_key will fail
    with pytest.raises(ValueError):
        _ = bundle.object_decryption_key


def test_recovery_bundle_get_dict(sample_recovery_bundle):
    swhid = ExtendedSWHID.from_string(
        "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645"
    )
    d = sample_recovery_bundle.get_dict(swhid)
    assert d == {"url": "https://github.com/user1/repo1"}


def test_recovery_bundle_write_content_data(tmp_path, sample_recovery_bundle):
    dest_path = tmp_path / "data"
    swhid = ExtendedSWHID.from_string(
        "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"
    )
    with open(dest_path, "wb") as dest:
        sample_recovery_bundle.write_content_data(swhid, dest)
    assert open(dest_path, "rb").read() == b"42\n"


def test_recovery_bundle_contents(sample_recovery_bundle, sample_data):
    contents = list(sample_recovery_bundle.contents())
    assert contents == [sample_data.content, sample_data.content3]


def test_recovery_bundle_skipped_contents(sample_recovery_bundle, sample_data):
    skipped_contents = list(sample_recovery_bundle.skipped_contents())
    assert skipped_contents == [sample_data.skipped_content]


def test_recovery_bundle_directories(sample_recovery_bundle, sample_data):
    directories = list(sample_recovery_bundle.directories())
    # Comparing Directory objects directly will not work because order
    # of entries are not guaranteed by storage. Therefore we have to
    # compare each field individually.
    assert len(directories) == 3
    for d, sample_d in zip(
        directories,
        [sample_data.directory, sample_data.directory5, sample_data.directory6],
    ):
        assert d.id == sample_d.id
        assert set(d.entries) == set(sample_d.entries)
        assert d.raw_manifest == sample_d.raw_manifest


def test_recovery_bundle_revisions(sample_recovery_bundle, sample_data):
    revisions = list(sample_recovery_bundle.revisions())
    assert revisions == [sample_data.revision, sample_data.revision2]


def test_recovery_bundle_releases(sample_recovery_bundle, sample_data):
    releases = list(sample_recovery_bundle.releases())
    assert releases == [sample_data.release, sample_data.release2]


def test_recovery_bundle_snapshots(sample_recovery_bundle, sample_data):
    snapshots = list(sample_recovery_bundle.snapshots())
    assert snapshots == [sample_data.snapshot, sample_data.complete_snapshot]


def test_recovery_bundle_origins(sample_recovery_bundle, sample_data):
    origins = list(sample_recovery_bundle.origins())
    assert origins == [sample_data.origin, sample_data.origin2]


def test_recovery_bundle_origin_visits(sample_recovery_bundle, sample_data):
    origin_visits = list(sample_recovery_bundle.origin_visits(sample_data.origin))
    assert origin_visits == [
        sample_data.origin_visit,
        sample_data.origin_visit2,
    ]


def test_recovery_bundle_origin_visit_statuses(sample_recovery_bundle, sample_data):
    origin_visit_statuses = list(
        sample_recovery_bundle.origin_visit_statuses(sample_data.origin)
    )
    assert origin_visit_statuses == [
        sample_data.origin_visit_status,
        sample_data.origin_visit2_status,
    ]


def test_restore(sample_recovery_bundle, swh_storage, sample_data):
    result = sample_recovery_bundle.restore(swh_storage)
    assert result == {
        "content:add": 2,
        "content:add:bytes": 10,
        "skipped_content:add": 1,
        "directory:add": 3,
        "revision:add": 2,
        "release:add": 2,
        "snapshot:add": 2,
        "origin:add": 2,
    }
    [
        origin,
    ] = swh_storage.origin_get(["https://github.com/user1/repo1"])
    assert origin is not None
    assert str(origin.swhid()) == "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645"
    origin_visit = swh_storage.origin_visit_get_by(origin.url, 1)
    assert origin_visit == sample_data.origin_visit
    origin_visit2 = swh_storage.origin_visit_get_by(origin.url, 2)
    assert origin_visit2 == sample_data.origin_visit2
    origin_visit2_status = swh_storage.origin_visit_status_get_latest(origin.url, 2)
    assert origin_visit2_status == sample_data.origin_visit2_status


def test_rollover(
    tmp_path,
    sample_recovery_bundle_path,
    secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each,
):
    import shutil

    bundle_path = shutil.copy(
        sample_recovery_bundle_path, tmp_path / "rollover.swh-recovery-bundle"
    )
    secret_sharing = (
        secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each
    )
    bundle = RecoveryBundle(bundle_path, object_decryption_key_provider_for_sample)

    # Record OriginVisit and OriginVisitStatuses objects for later comparison
    origin_visits = set()
    origin_visit_statuses = set()
    for origin in bundle.origins():
        origin_visits.update(bundle.origin_visits(origin))
        origin_visit_statuses.update(bundle.origin_visit_statuses(origin))

    # Perform rollover!
    bundle.rollover(secret_sharing)

    # Check all the things!!
    def decryption_key_provider(manifest):
        return recover_object_decryption_key_from_encrypted_shares(
            manifest.decryption_key_shares,
            available_secret_keys_for_2_groups_required_of_3_one_and_two,
        )

    new_bundle = RecoveryBundle(bundle_path, decryption_key_provider)
    # Do we have the new holders?
    assert new_bundle.share_ids == {"Ali", "Bob", "Camille", "Dlique", "Essun"}
    # Has the old object been updated as well?
    assert bundle.share_ids == new_bundle.share_ids
    # Is the decryption key still the same after being recovered?
    assert new_bundle.object_decryption_key == OBJECT_SECRET_KEY
    # Can we still decrypt all known objects?
    decrypted_swhids = set()
    for obj in itertools.chain(
        new_bundle.contents(),
        new_bundle.skipped_contents(),
        new_bundle.directories(),
        new_bundle.revisions(),
        new_bundle.releases(),
        new_bundle.snapshots(),
    ):
        decrypted_swhids.add(str(obj.swhid()))
    new_origin_visits = set()
    new_origin_visit_statuses = set()
    for origin in new_bundle.origins():
        decrypted_swhids.add(str(origin.swhid()))
        new_origin_visits.update(new_bundle.origin_visits(origin))
        new_origin_visit_statuses.update(new_bundle.origin_visit_statuses(origin))
    assert decrypted_swhids == set(bundle.swhids)
    assert origin_visits == new_origin_visits
    assert origin_visit_statuses == new_origin_visit_statuses


def test_rollover_fails_when_unable_to_write(
    tmp_path,
    sample_recovery_bundle_path,
    secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each,
):
    import shutil

    bundle_path = shutil.copy(
        sample_recovery_bundle_path, tmp_path / "rollover.swh-recovery-bundle"
    )
    secret_sharing = (
        secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each
    )
    bundle = RecoveryBundle(bundle_path, object_decryption_key_provider_for_sample)
    share_ids = bundle.share_ids
    swhids = bundle.swhids

    # We can more or less simulate a full disk by preventing writes
    # in the directory holding the recovery bundle.
    os.chmod(tmp_path, 0o500)

    with pytest.raises(OSError):
        bundle.rollover(secret_sharing)

    new_bundle = RecoveryBundle(bundle_path, object_decryption_key_provider_for_sample)
    assert new_bundle.share_ids == share_ids
    assert new_bundle.object_decryption_key == OBJECT_SECRET_KEY
    assert new_bundle.get_dict(ExtendedSWHID.from_string(swhids[0])) is not None



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
This module implements the marking stage of the
:ref:`removal algorithm <alter_removal_algorithm>`.
"""

from enum import Enum, auto
from itertools import chain
import logging
from typing import Iterator, List

from igraph import Vertex

from swh.graph.http_client import GraphArgumentException, RemoteGraphClient
from swh.model.swhids import ExtendedObjectType as ObjectType
from swh.model.swhids import ExtendedSWHID
from swh.storage.interface import StorageInterface

from .inventory import InventorySubgraph
from .subgraph import Subgraph

logger = logging.getLogger(__name__)


class MarkingState(Enum):
    """Represents the different state of vertices while the marking algorithm is on-going.

    All vertices start UNMARKED. After examination of their inbound references, they are
    either marked as REMOVABLE or UNREMOVABLE.

    :meta private:
    """

    UNMARKED = auto()
    REMOVABLE = auto()
    UNREMOVABLE = auto()


class RemovableSubgraph(Subgraph):
    """A class representing the subgraph of objects that can be safely removed
    from the archive

    This subgraph should only be created from an :py:class:`InventorySubgraph`
    using :py:meth:`from_inventory_subgraph`. No new vertices or edges can be
    added at this point.

    Getting the list of SWHID that can be removed is done by using
    py:meth:`removable_swhids`.

    After marking happened, this subgraph will contain all candidates until
    :py:meth:`delete_unremovable` is called.
    """

    def __init__(self, *args, **kwargs):
        """ """
        super().__init__(*args, **kwargs)
        self["name"] = "Removable"
        self.vs["state"] = MarkingState.UNMARKED

    @classmethod
    def from_inventory_subgraph(cls, subgraph: InventorySubgraph):
        return cls.copy(subgraph)

    def add_vertex(self, name: str, **kwargs):
        """Not available for RemovableSubgraph.

        :meta private:"""
        raise NotImplementedError("No new vertex should be added at this stage.")

    def add_edge(self, src: Vertex, dst: Vertex, skip_duplicates=False, **kwargs):
        """Not available for RemovableSubgraph.

        :meta private:"""
        raise NotImplementedError("No new edge should be added at this stage.")

    def delete_unremovable(self):
        """Delete all vertices for unremovable objects from this subgraph."""

        self.delete_vertices(self.vs.select(state_eq=MarkingState.UNREMOVABLE))

    def removable_swhids(self) -> List[ExtendedSWHID]:
        """Returns a list of SWHIDs that can safely be removed from the archive."""

        return [
            v["swhid"] for v in self.select_ordered(state_eq=MarkingState.REMOVABLE)
        ]

    def dot_node_attributes(self, v: Vertex) -> List[str]:
        """Get a list of attributes in DOT format for the given vertex.

        On top of default attributes, color the background of each node if
        they can be removed or not.

        :meta private:
        """
        attrs = super().dot_node_attributes(v)
        # Unset the shape so we can better spot ExtendedSWHID that can be removed
        if v["state"] == MarkingState.UNMARKED:
            attrs.append('style="filled,dotted"')
        elif v["state"] == MarkingState.UNREMOVABLE:
            attrs.append("fillcolor=white")
        else:
            attrs.append("fillcolor=red")
        return attrs


class Marker:
    """A class encapsulating our algorithm marking nodes as removable or not.

    :meta private:
    """

    def __init__(
        self,
        storage: StorageInterface,
        graph_client: RemoteGraphClient,
        subgraph: RemovableSubgraph,
    ):
        self._storage = storage
        self._graph_client = graph_client
        self._subgraph = subgraph
        if logger.isEnabledFor(logging.INFO):
            self._total_nodes = len(subgraph.vs)
            self._marked_count = 0

    @property
    def subgraph(self):
        return self._subgraph

    def has_unknown_inbound_edges(self, vertex: Vertex):
        # Origins never have any inbound edges
        if vertex["swhid"].object_type == ObjectType.ORIGIN:
            return False
        # We use a set of str as an optimization because
        # swh.graph API will return SWHIDs as str
        known_predecessors = {str(pred["swhid"]) for pred in vertex.predecessors()}
        # We only need to find one extra edge from the ones we already know
        search_limit = len(known_predecessors) + 1
        for pred in chain(
            self.inbound_edges_from_graph(vertex, search_limit),
            self.inbound_edges_from_storage(vertex, search_limit),
        ):
            # Removing a revision used as a submodule should be possible.
            # (We do not care that it will make submodule users miss part
            # of their source code.)
            # We use string matching here because the graph API returns
            # str objects and not ExtendedSWHID objects.
            if vertex["swhid"].object_type == ObjectType.REVISION and pred.startswith(
                "swh:1:dir:"
            ):
                logger.info(
                    "Skipping predecessor %s of %s as its a submodule",
                    pred,
                    vertex["swhid"],
                )
                return False
            logger.debug("Is %s an extra predecessor of %s?", pred, vertex["swhid"])
            if pred not in known_predecessors:
                return True
        return False

    def inbound_edges_from_graph(
        self, vertex: Vertex, search_limit: int
    ) -> Iterator[str]:
        try:
            yield from self._graph_client.neighbors(
                str(vertex["swhid"]), direction="backward", max_edges=search_limit
            )
        except GraphArgumentException:
            yield from ()

    def inbound_edges_from_storage(
        self, vertex: Vertex, search_limit: int
    ) -> Iterator[str]:
        yield from map(
            str,
            self._storage.object_find_recent_references(
                vertex["swhid"], limit=search_limit
            ),
        )

    def mark_candidates(self):
        for index, vid in enumerate(self._subgraph.topological_sorting()):
            vertex = self._subgraph.vs[vid]
            assert all(
                pred["state"] != MarkingState.UNMARKED for pred in vertex.predecessors()
            ), "topological sort broken: one predecessor is still in unmarked state"
            # If any predecessor is unremovable, this makes this object unremovable too
            if any(
                pred["state"] == MarkingState.UNREMOVABLE
                for pred in vertex.predecessors()
            ):
                vertex["state"] = MarkingState.UNREMOVABLE
                continue
            vertex["state"] = (
                MarkingState.UNREMOVABLE
                if self.has_unknown_inbound_edges(vertex)
                else MarkingState.REMOVABLE
            )
            self.log_marking_progress()

    def log_marking_progress(self):
        if not logger.isEnabledFor(logging.INFO):
            return
        self._marked_count += 1
        if self._total_nodes == self._marked_count or self._marked_count % 100 == 0:
            logger.info(
                "Checking inbound edges for node %s/%s",
                self._marked_count,
                self._total_nodes,
            )


def mark_removable(
    storage: StorageInterface,
    graph_client: RemoteGraphClient,
    inventory_subgraph: InventorySubgraph,
) -> RemovableSubgraph:
    """Find which candidates can be safely removed from the archive.

    Using the previously populated ``inventory_subgraph``, query the given
    ``storage`` and ``graph_client`` to assess inbound references for each
    candidate and see which one can be safely removed from the archive.
    """
    subgraph = RemovableSubgraph.from_inventory_subgraph(inventory_subgraph)
    marker = Marker(storage, graph_client, subgraph)
    marker.mark_candidates()
    return marker.subgraph



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Set

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

if TYPE_CHECKING:
    from swh.model.swhids import ExtendedSWHID

    from .recovery_bundle import ObjectDecryptionKeyProvider, ShareDecryptionKeys


class SwhidOrUrlParamType(click.ParamType):
    name = "swhid or origin URL"

    def convert(self, value, param, ctx):
        import hashlib

        from swh.model.exceptions import ValidationError
        from swh.model.swhids import ExtendedSWHID

        if value.startswith("swh:1:"):
            try:
                return ExtendedSWHID.from_string(value)
            except ValidationError:
                self.fail(f"expected extended SWHID, got {value!r}", param, ctx)
        else:
            click.secho(f"Assuming {value} is an origin URL.", fg="cyan")
            sha1 = hashlib.sha1(value.encode("utf-8")).hexdigest()
            swhid = ExtendedSWHID.from_string(f"swh:1:ori:{sha1}")
            return swhid


DEFAULT_CONFIG = {
    "storage": {
        "cls": "postgresql",
        "db": "dbname=softwareheritage host=db.internal.softwareheritage.org user=guest",
        "objstorage": {
            "cls": "memory",
        },
    },
    "graph_client": {
        "url": "http://granet.internal.softwareheritage.org:5009",
        # timeout is in seconds
        # see https://requests.readthedocs.io/en/latest/user/advanced/#timeouts
        "timeout": 10,
    },
    "recovery_bundles": {
        "secret_sharing": {
            "minimum_required_groups": 2,
            "groups": {},
        }
    },
}


@swh_cli_group.group(name="alter", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def alter_cli_group(ctx):
    """Archive alteration tools.

    Location of the configuration should be specified through the environment
    variable ``SWH_CONFIG_FILENAME``.

    Expected config format:

        \b
        storage:
          cls: postgresql
          db: "service=…"
          objstorage:
            cls: memory
        \b
        graph:
          cls: remote
          url: "http://granet.internal.softwareheritage.org:5009"
        \b
        recovery_bundles:
          secret_sharing:
            minimum_required_groups: 2
            groups:
              legal:
                minimum_required_shares: 1
                recipient_keys:
                    "YubiKey serial 4245067 slot 1": age1yubikey1q2e37f74zzazz75mtggzql3at66pegemfnul0dtd7axctahljkvsqezscaq
                    "YubiKey serial 2284622 slot 3": age1yubikey4o1aypv83isatti92q1zasv1hkpuozlkoak4zd66t7poud23rftqrcszjgul
              sysadmins:
                minimum_required_shares: 1
                recipient_keys:
                    "YubiKey serial 3862152 slot 1": age1yubikeyrupnxsu6uneqxw146g9szaofyxexiy4nhnzqg1ayb9b85g8h4oardwj6c212
                    "Ruby": age1y6epp27nq8n4faj8g8hkw8thcvj744y5vnr8jyfmp4857d6npc3qn9k7jz

    The identifier for the recipient key must be in the form of
    “YubiKey serial ####### slot #” if the secret key is stored
    on a YubiKey. Keys specified by any other identifiers will be
    considered as plain age identities.
    """  # noqa: B950
    from swh.core import config

    conf = config.load_from_envvar(default_config=DEFAULT_CONFIG)
    ctx.ensure_object(dict)
    ctx.obj["config"] = conf
    return ctx


@alter_cli_group.command()
@click.option(
    "--dry-run", is_flag=True, help="perform a trial run with no changes made"
)
@click.option(
    "--output-inventory-subgraph",
    type=click.File(mode="w", atomic=True),
)
@click.option(
    "--output-removable-subgraph",
    type=click.File(mode="w", atomic=True),
)
@click.option(
    "--output-pruned-removable-subgraph",
    type=click.File(mode="w", atomic=True),
)
@click.option(
    "--identifier",
    metavar="IDENTIFIER",
    required=True,
    help="identifier for this removal operation",
)
@click.option(
    "--reason",
    metavar="REASON",
    help="reason for this removal operation",
)
@click.option(
    "--expire",
    metavar="YYYY-MM-DD",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="date when the recovery bundle should be removed",
)
@click.option(
    "--recovery-bundle",
    metavar="PATH",
    type=click.Path(writable=True, dir_okay=False),
    required=True,
    help="path to the recovery bundle that will be created",
)
@click.argument(
    "swhids",
    metavar="<SWHID|URL>..",
    type=SwhidOrUrlParamType(),
    required=True,
    nargs=-1,
)
@click.pass_context
def remove(
    ctx,
    swhids: List["ExtendedSWHID"],
    dry_run: bool,
    output_inventory_subgraph,
    output_removable_subgraph,
    output_pruned_removable_subgraph,
    identifier,
    reason,
    expire,
    recovery_bundle,
) -> None:
    """Remove the given SWHIDs or URLs from the archive."""
    from swh.graph.http_client import RemoteGraphClient
    from swh.storage import get_storage

    from .inventory import make_inventory
    from .removable import mark_removable

    conf = ctx.obj["config"]
    storage = get_storage(**conf["storage"])
    graph_client = RemoteGraphClient(**conf["graph_client"])

    click.echo("Removing the following origins:")
    for swhid in swhids:
        click.echo(f" - {swhid}")
    click.secho("Inventorying all reachable objects…", fg="cyan")
    inventory_subgraph = make_inventory(storage, graph_client, swhids)
    if output_inventory_subgraph:
        inventory_subgraph.write_dot(output_inventory_subgraph)
        output_inventory_subgraph.close()
    click.secho("Determining which objects can be safely removed…", fg="cyan")
    removable_subgraph = mark_removable(storage, graph_client, inventory_subgraph)
    if output_removable_subgraph:
        removable_subgraph.write_dot(output_removable_subgraph)
        output_removable_subgraph.close()
    removable_subgraph.delete_unremovable()
    if output_pruned_removable_subgraph:
        removable_subgraph.write_dot(output_pruned_removable_subgraph)
        output_pruned_removable_subgraph.close()
    removable_swhids = list(removable_subgraph.removable_swhids())
    if dry_run:
        click.echo(f"We would remove {len(removable_swhids)} objects:")
        for swhid in removable_swhids:
            click.echo(f" - {swhid}")
        return
    click.confirm(
        click.style(
            f"Proceed with removing {len(removable_swhids)} objects?",
            fg="yellow",
            bold=True,
        ),
        abort=True,
    )

    from .recovery_bundle import (
        RecoveryBundleCreator,
        SecretSharing,
        generate_age_keypair,
    )

    object_public_key, object_secret_key = generate_age_keypair()
    secret_sharing = SecretSharing.from_dict(conf["recovery_bundles"]["secret_sharing"])
    decryption_key_shares = secret_sharing.generate_encrypted_shares(
        identifier, object_secret_key
    )
    click.secho("Creating recovery bundle…", fg="cyan")
    with RecoveryBundleCreator(
        path=recovery_bundle,
        storage=storage,
        removal_identifier=identifier,
        object_public_key=object_public_key,
        decryption_key_shares=decryption_key_shares,
    ) as creator:
        if reason is not None:
            creator.set_reason(reason)
        if expire is not None:
            try:
                creator.set_expire(expire.astimezone())
            except ValueError as ex:
                click.echo(f"Invalid expiration date: {str(ex)}", err=True)
                ctx.exit(1)
        creator.backup_swhids(removable_swhids)
    click.secho("Recovery bundle created.", fg="green")
    raise NotImplementedError("Actual removal still need to be written")


@alter_cli_group.group(name="recovery-bundle", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def recovery_bundle_cli_group(ctx):
    """Recovery bundle related tools."""
    return ctx


@recovery_bundle_cli_group.command(name="info")
@click.option(
    "--dump-manifest",
    is_flag=True,
    default=False,
    help="Show raw manifest in YAML format.",
)
@click.option(
    "--show-encrypted-secrets",
    is_flag=True,
    default=False,
    help="Show encrypted secrets.",
)
@click.argument(
    "recovery-bundle",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=True,
)
@click.pass_context
def info(ctx, recovery_bundle, dump_manifest, show_encrypted_secrets) -> None:
    """Display the manifest of the given recovery bundle."""
    from .recovery_bundle import RecoveryBundle

    bundle = RecoveryBundle(recovery_bundle)

    if dump_manifest:
        click.echo(bundle.dump_manifest(), nl=False)
        ctx.exit()

    title = f"Recovery bundle “{bundle.removal_identifier}”"
    click.echo(title)
    click.echo("=" * len(title))
    click.echo("")
    click.echo(f"Created: {bundle.created.isoformat()}")
    if bundle.reason:
        lines = bundle.reason.rstrip().split("\n")
        lines[0] = f"Reason: {lines[0]}"
        click.echo("\n        ".join(lines))
    if bundle.expire:
        click.echo(f"Expire: {bundle.expire}")
    click.echo("List of SWHID objects:")
    for swhid in bundle.swhids:
        click.echo(f"- {swhid}")
    click.echo("Secret share holders:")
    for share_id in sorted(bundle.share_ids):
        click.echo(f"- {share_id}")
        if show_encrypted_secrets:
            click.echo(bundle.encrypted_secret(share_id))


def _share_decryption_keys_provider(share_ids: Set[str]) -> ShareDecryptionKeys:
    import subprocess
    import sys

    from .recovery_bundle import list_yubikey_identities

    for attempt in range(1, 10):
        if not any(share_id.startswith("YubiKey") for share_id in share_ids):
            # No shares require a YubiKey, so there is nothing we can do here
            break
        try:
            for share_id, secret_key in list_yubikey_identities():
                if share_id not in share_ids:
                    continue
                share_ids.remove(share_id)
                click.echo(
                    "🔧 Decrypting share using "
                    f"{click.style(share_id, fg='magenta', bold=True)}…"
                )
                click.echo("💭 You might need to tap the right YubiKey when it blinks.")
                yield share_id, secret_key
                click.echo()
        except subprocess.CalledProcessError as ex:
            if "age-plugin-yubikey" not in ex.cmd[0]:
                raise
            click.echo(
                f"""💥 {click.style('age-plugin-yubikey failed to '
                                    'list connected YubiKeys.', bold=True, fg='red')}"""
            )
            click.echo("💭 Please disconnect all YubiKeys and retry.")
            sys.exit(1)
        if share_ids:
            yubikey_ids = list(sorted(share_ids))
            if len(yubikey_ids) > 1:
                yubikeys = ", ".join(
                    click.style(share_id, fg="magenta", bold=True)
                    for share_id in yubikey_ids[:-1]
                )
                yubikeys += " or " + click.style(
                    yubikey_ids[-1], fg="magenta", bold=True
                )
            else:
                yubikeys = click.style(yubikey_ids[0], fg="magenta", bold=True)
            click.prompt(
                f"🔐 Please insert {yubikeys} and press "
                f"{click.style('Enter', fg='green', bold=True)}…",
                default="Ok",
                show_default=False,
                hide_input=True,
                prompt_suffix="",
            )
    click.echo(
        f"""💥 {click.style('Unable to decrypt enough shared secrets to recover '
                                   'the object decryption key. Aborting.',
                                   bold=True, fg='red')}"""
    )
    sys.exit(1)


def _print_decrypted_mnemonic(mnemonic: str, share_id: Optional[str] = None) -> None:
    fmt_from = ""
    if share_id:
        fmt_from = f" from {click.style(share_id, fg='magenta', bold=True)}"
    click.echo(f"🔑 Recovered shared secret{fmt_from}:")
    # Quoting from SLIP-0039: This construction yields a beneficial
    # property where the random identifier and the iteration exponent
    # transform into the first two words of the mnemonic code, so the user
    # can immediately tell whether the correct shares are being combined,
    # i.e. they have to have the same first two words. Moreover, the third
    # word encodes the group index, group threshold and part of the group
    # count. Since the group threshold and group count are constant, all
    # **shares belonging to the same group start with the same three words**.
    words = mnemonic.split()
    click.echo(
        " ".join(
            click.style(word, fg="blue", bold=index < 3)
            for index, word in enumerate(words)
        )
    )


def _recover_mnemonics_from_identity_files(
    manifest, share_ids, identity_files, show_decrypted_mnemonics
):
    from .recovery_bundle import WrongDecryptionKey, age_decrypt_from_identity

    # As we can’t know which identity file corresponds to which encrypted shared
    # secret, we have to try them all and see which one we can actually decrypt.
    recovered = {}
    for identity_file in identity_files:
        for share_id in share_ids:
            try:
                recovered[share_id] = age_decrypt_from_identity(
                    identity_file, manifest.decryption_key_shares[share_id]
                ).decode("us-ascii")
                if show_decrypted_mnemonics:
                    _print_decrypted_mnemonic(recovered[share_id], share_id)
            except WrongDecryptionKey:
                pass
    return recovered


def prompting_object_decryption_key_provider(
    manifest, known_mnemonics=None, identity_files=None, show_decrypted_mnemonics=False
) -> str:
    import functools

    from .recovery_bundle import recover_object_decryption_key_from_encrypted_shares

    decrypted_mnemonic_processor = None
    if show_decrypted_mnemonics:
        decrypted_mnemonic_processor = _print_decrypted_mnemonic
    share_ids = set(manifest.decryption_key_shares.keys())
    # Normalize known_mnemonics
    known_mnemonics = list(known_mnemonics or [])
    if identity_files:
        recovered = _recover_mnemonics_from_identity_files(
            manifest, share_ids, identity_files, show_decrypted_mnemonics
        )
        share_ids.difference_update(recovered.keys())
        known_mnemonics.extend(recovered.values())
    yubikey_share_ids = set(
        share_id for share_id in share_ids if share_id.startswith("YubiKey")
    )
    missing_ids = share_ids - yubikey_share_ids
    if missing_ids:
        fmt_ids = ", ".join(
            click.style(share_id, fg="magenta", bold=True) for share_id in missing_ids
        )
        click.echo(
            f"""\n🚸 {click.style('The following secret shares will not be '
                                    'decrypted:', fg='yellow')} {fmt_ids}\n"""
        )

    return recover_object_decryption_key_from_encrypted_shares(
        manifest.decryption_key_shares,
        functools.partial(_share_decryption_keys_provider, yubikey_share_ids),
        decrypted_mnemonic_processor=decrypted_mnemonic_processor,
        known_mnemonics=known_mnemonics,
    )


def get_object_decryption_key_provider(ctx) -> ObjectDecryptionKeyProvider:
    import functools

    secrets = ctx.params.get("secret")
    identity_files = ctx.params.get("identity")
    object_decryption_key_provider: ObjectDecryptionKeyProvider = functools.partial(
        prompting_object_decryption_key_provider,
        known_mnemonics=secrets,
        identity_files=identity_files,
    )
    decryption_key = ctx.params.get("decryption_key")
    if decryption_key:
        if not decryption_key.lower().startswith("age-secret-key-"):
            ctx.fail(
                "The given decryption key does not look like a decryption key. "
                "It should start with “AGE-SECRET-KEY-”"
            )

        def known_key_provider(_):
            return decryption_key

        object_decryption_key_provider = known_key_provider
    return object_decryption_key_provider


class ContentSWHID(click.ParamType):
    name = "swhid of a content object"

    def convert(self, value, param, ctx):
        from swh.model.swhids import ExtendedObjectType, ExtendedSWHID, ValidationError

        try:
            swhid = ExtendedSWHID.from_string(value)
        except ValidationError:
            self.fail(f"expected SWHID, got {value!r}", param, ctx)
        if swhid.object_type != ExtendedObjectType.CONTENT:
            self.fail("We can only extract data for Content objects", param, ctx)
        return swhid


@recovery_bundle_cli_group.command(name="extract-content")
@click.option(
    "-o",
    "--output",
    type=click.File("wb"),
    metavar="FILE",
    required=True,
    help="write data to FILE",
)
@click.option(
    "--decryption-key",
    metavar="AGE_SECRET_KEY",
    help="use the given decryption key instead of the bundle shared secrets",
)
@click.option(
    "-s",
    "--secret",
    metavar="MNEMONIC",
    multiple=True,
    help="Known shared secret. May be repeated.",
)
@click.option(
    "-i",
    "--identity",
    metavar="IDENTITY",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    multiple=True,
    help="Path to file with age identities. May be repeated.",
)
@click.argument(
    "recovery-bundle",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=True,
)
@click.argument(
    "SWHID",
    type=ContentSWHID(),
    required=True,
)
@click.pass_context
def extract_content(
    ctx,
    output,
    recovery_bundle,
    swhid,
    decryption_key=None,
    identity=None,
    secret=None,
) -> None:
    """Extract data from content stored in a recovery bundle."""
    from .recovery_bundle import RecoveryBundle, WrongDecryptionKey

    secret_key_provider = get_object_decryption_key_provider(ctx)
    bundle = RecoveryBundle(recovery_bundle, secret_key_provider)

    if str(swhid) not in bundle.swhids:
        click.secho(
            f"“{swhid}” is not in the recovery bundle", err=True, fg="red", bold=True
        )
        ctx.exit(1)

    try:
        bundle.write_content_data(swhid, output)
    except WrongDecryptionKey:
        click.secho(
            f"Wrong decryption key for this bundle ({bundle.removal_identifier})",
            err=True,
            fg="red",
            bold=True,
        )
        ctx.exit(2)


@recovery_bundle_cli_group.command(name="restore")
@click.option(
    "--decryption-key",
    metavar="AGE_SECRET_KEY",
    help="use the given decryption key instead of the bundle shared secrets",
)
@click.option(
    "-s",
    "--secret",
    metavar="MNEMONIC",
    multiple=True,
    help="Known shared secret. May be repeated.",
)
@click.option(
    "-i",
    "--identity",
    metavar="IDENTITY",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    multiple=True,
    help="Path to file with age identities. May be repeated.",
)
@click.argument(
    "recovery-bundle",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=True,
)
@click.pass_context
def restore(
    ctx, recovery_bundle, decryption_key=None, identity=None, secret=None
) -> None:
    """Restore a recovery bundle to Software Heritage archive."""
    from .recovery_bundle import RecoveryBundle, WrongDecryptionKey

    conf = ctx.obj["config"]
    from swh.storage import get_storage

    storage = get_storage(**conf["storage"])

    secret_key_provider = get_object_decryption_key_provider(ctx)
    bundle = RecoveryBundle(recovery_bundle, secret_key_provider)
    try:
        # XXX: we could use click.progressbar here
        results = bundle.restore(storage)
    except WrongDecryptionKey:
        click.echo(
            f"Wrong decryption key for this bundle ({bundle.removal_identifier})"
        )
        ctx.exit(2)

    click.echo(
        "Restoration complete. Results: \n"
        f"- Content objects added: {results['content:add']}\n"
        f"- Total bytes added to objstorage: {results['content:add:bytes']}\n"
        f"- SkippedContent objects added: {results['skipped_content:add']}\n"
        f"- Directory objects added: {results['directory:add']}\n"
        f"- Revision objects added: {results['revision:add']}\n"
        f"- Release objects added: {results['release:add']}\n"
        f"- Snapshot objects added: {results['snapshot:add']}\n"
        f"- Origin objects added: {results['origin:add']}\n"
    )


def _strip_rage_report(output):
    # rage prompts for report when it errors like this:
    #   [ Did rage not do what you expected? Could an error be more useful? ]
    #   [ Tell us: https://str4d.xyz/rage/report                            ]
    # This can be confusing in our case so strip them from the output.
    return b"\n".join(
        line
        for line in output.split(b"\n")
        if not line.startswith(b"[") and not line.endswith(b"]")
    )


@recovery_bundle_cli_group.command(name="recover-decryption-key")
@click.option(
    "-s",
    "--secret",
    metavar="MNEMONIC",
    multiple=True,
    help="Known shared secret. May be repeated.",
)
@click.option(
    "-i",
    "--identity",
    metavar="IDENTITY",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    multiple=True,
    help="Path to file with age identities. May be repeated.",
)
@click.option(
    "--show-recovered-secrets",
    is_flag=True,
    default=False,
    help="Show recovered shared secrets. Useful for remote/distributed recoveries.",
)
@click.argument(
    "recovery-bundle",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=True,
)
def recover_decryption_key(
    recovery_bundle, secret, identity, show_recovered_secrets
) -> None:
    """Recover the decryption key using shared secrets."""
    import subprocess
    import sys

    from .recovery_bundle import RecoveryBundle

    def object_decryption_key_provider(*args, **kwargs):
        kwargs["known_mnemonics"] = list(secret)
        kwargs["identity_files"] = list(identity)
        kwargs["show_decrypted_mnemonics"] = show_recovered_secrets
        return prompting_object_decryption_key_provider(*args, **kwargs)

    try:
        bundle = RecoveryBundle(recovery_bundle, object_decryption_key_provider)
        decryption_key = bundle.object_decryption_key
        click.echo(
            f"\n🔓 Recovered decryption key:\n{click.style(decryption_key, bold=True)}"
        )
    except subprocess.CalledProcessError as ex:
        if "rage" not in ex.cmd[0] and ex.cmd[1] != "--decrypt":
            raise
        click.echo(
            f"""💥 {click.style('rage decryption failed:', bold=True, fg='red')}"""
        )
        click.echo(_strip_rage_report(ex.stderr))
        sys.exit(1)


@recovery_bundle_cli_group.command(name="rollover")
@click.option(
    "--decryption-key",
    metavar="AGE_SECRET_KEY",
    help="use the given decryption key instead of the bundle shared secrets",
)
@click.option(
    "-s",
    "--secret",
    metavar="MNEMONIC",
    multiple=True,
    help="Known shared secret. May be repeated.",
)
@click.option(
    "-i",
    "--identity",
    metavar="IDENTITY",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    multiple=True,
    help="Path to file with age identities. May be repeated.",
)
@click.argument(
    "recovery-bundles",
    metavar="[RECOVERY_BUNDLE]…",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    nargs=-1,
)
@click.pass_context
def rollover(
    ctx, recovery_bundles, decryption_key=None, identity=None, secret=None
) -> None:
    """Rollover recovery bundles to new shared secrets."""
    conf = ctx.obj["config"]

    from .recovery_bundle import RecoveryBundle, SecretSharing, WrongDecryptionKey

    secret_key_provider = get_object_decryption_key_provider(ctx)
    secret_sharing = SecretSharing.from_dict(conf["recovery_bundles"]["secret_sharing"])
    click.secho("New shared secret holders:")
    for share_id in sorted(secret_sharing.share_ids):
        click.echo(f"- {click.style(share_id, fg='magenta', bold=True)}")
    click.confirm(
        click.style(
            "Proceed with rolling over the shared secrets?",
            fg="yellow",
            bold=True,
        ),
        abort=True,
    )
    for recovery_bundle in recovery_bundles:
        bundle = RecoveryBundle(recovery_bundle, secret_key_provider)
        # Ensure that we can decrypt at least some objects with the provided key
        try:
            origin = list(bundle.origins())
            assert len(origin) > 0, "Oops! No Origin objects in this recovery bundle."
        except WrongDecryptionKey:
            click.secho(
                f"Wrong decryption key for this bundle ({bundle.removal_identifier})",
                err=True,
                fg="red",
                bold=True,
            )
            ctx.exit(2)
        bundle.rollover(secret_sharing)
        click.secho("Shared secrets for ", fg="green", nl=False)
        click.secho(bundle.removal_identifier, fg="green", bold=True, nl=False)
        click.secho(" have been rolled over.", fg="green")



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict
from typing import List

import pytest

from swh.graph.example_dataset import FORKED_ORIGIN, INITIAL_ORIGIN
from swh.model.swhids import ExtendedSWHID

from ..inventory import InventorySubgraph
from ..removable import mark_removable
from .test_inventory import (  # noqa: F401
    directory_6_with_multiple_entries_pointing_to_the_same_content,
    snapshot_20_with_multiple_branches_pointing_to_the_same_head,
)
from .test_inventory import graph_client_with_both_origins  # noqa: F401
from .test_inventory import graph_client_with_only_initial_origin  # noqa: F401
from .test_inventory import graph_client_with_submodule  # noqa: F401
from .test_inventory import origin_with_submodule  # noqa: F401
from .test_inventory import sample_populated_storage  # noqa: F401
from .test_subgraph import write_dot_if_requested


@pytest.fixture
def inventory_from_initial_origin():
    g = InventorySubgraph()
    v_ori = g.add_swhid(INITIAL_ORIGIN.swhid())
    v_snp = g.add_swhid("swh:1:snp:0000000000000000000000000000000000000020")
    v_rel = g.add_swhid("swh:1:rel:0000000000000000000000000000000000000010")
    v_rev_09 = g.add_swhid("swh:1:rev:0000000000000000000000000000000000000009")
    v_rev_03 = g.add_swhid("swh:1:rev:0000000000000000000000000000000000000003")
    v_dir_08 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000008")
    v_dir_06 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000006")
    v_dir_02 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000002")
    v_cnt_07 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000007")
    v_cnt_05 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000005")
    v_cnt_04 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000004")
    v_cnt_01 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000001")
    g.add_edge(v_ori, v_snp)
    g.add_edge(v_snp, v_rel)
    g.add_edge(v_snp, v_rev_09)
    g.add_edge(v_rel, v_rev_09)
    g.add_edge(v_rev_09, v_rev_03)
    g.add_edge(v_rev_09, v_dir_08)
    g.add_edge(v_rev_03, v_dir_02)
    g.add_edge(v_dir_08, v_dir_06)
    g.add_edge(v_dir_08, v_cnt_07)
    g.add_edge(v_dir_08, v_cnt_01)
    g.add_edge(v_dir_06, v_cnt_05)
    g.add_edge(v_dir_06, v_cnt_04)
    g.add_edge(v_dir_02, v_cnt_01)
    write_dot_if_requested(g, "inventory_from_initial_origin.dot")
    return g


@pytest.fixture
def inventory_from_forked_origin():
    g = InventorySubgraph()
    v_ori = g.add_swhid(FORKED_ORIGIN.swhid())
    v_snp = g.add_swhid("swh:1:snp:0000000000000000000000000000000000000022")
    v_rel_21 = g.add_swhid("swh:1:rel:0000000000000000000000000000000000000021")
    v_rel_10 = g.add_swhid("swh:1:rel:0000000000000000000000000000000000000010")
    v_rev_18 = g.add_swhid("swh:1:rev:0000000000000000000000000000000000000018")
    v_rev_13 = g.add_swhid("swh:1:rev:0000000000000000000000000000000000000013")
    v_rev_09 = g.add_swhid("swh:1:rev:0000000000000000000000000000000000000009")
    v_rev_03 = g.add_swhid("swh:1:rev:0000000000000000000000000000000000000003")
    v_dir_17 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000017")
    v_dir_16 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000016")
    v_dir_12 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000012")
    v_dir_08 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000008")
    v_dir_06 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000006")
    v_dir_02 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000002")
    v_cnt_15 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000015")
    v_cnt_14 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000014")
    v_cnt_11 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000011")
    v_cnt_07 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000007")
    v_cnt_05 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000005")
    v_cnt_04 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000004")
    v_cnt_01 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000001")
    g.add_edge(v_ori, v_snp)
    g.add_edge(v_snp, v_rel_21)
    g.add_edge(v_snp, v_rel_10)
    g.add_edge(v_snp, v_rev_09)
    g.add_edge(v_rel_21, v_rev_18)
    g.add_edge(v_rev_18, v_rev_13)
    g.add_edge(v_rev_13, v_rev_09)
    g.add_edge(v_rev_09, v_rev_03)
    g.add_edge(v_rev_18, v_dir_17)
    g.add_edge(v_rev_13, v_dir_12)
    g.add_edge(v_rev_09, v_dir_08)
    g.add_edge(v_rev_03, v_dir_02)
    g.add_edge(v_dir_17, v_dir_16)
    g.add_edge(v_dir_12, v_dir_08)
    g.add_edge(v_dir_08, v_dir_06)
    g.add_edge(v_dir_17, v_cnt_14)
    g.add_edge(v_dir_16, v_cnt_15)
    g.add_edge(v_dir_12, v_cnt_11)
    g.add_edge(v_dir_08, v_cnt_07)
    g.add_edge(v_dir_08, v_cnt_01)
    g.add_edge(v_dir_06, v_cnt_05)
    g.add_edge(v_dir_06, v_cnt_04)
    g.add_edge(v_dir_02, v_cnt_01)
    write_dot_if_requested(g, "inventory_from_forked_origin.dot")
    return g


@pytest.fixture
def storage_with_no_new_references_since_export(
    mocker, sample_populated_storage  # noqa: F811
):
    # Recent API, see:
    # https://gitlab.softwareheritage.org/swh/devel/swh-storage/-/merge_requests/1042
    mocker.patch.object(
        sample_populated_storage,
        "object_find_recent_references",
        create=True,
        return_value=[],
    )
    return sample_populated_storage


@pytest.fixture
def storage_with_references_from_forked_origin(
    mocker, sample_populated_storage, inventory_from_forked_origin  # noqa: F811
):
    references = defaultdict(list)
    for e in inventory_from_forked_origin.es:
        references[inventory_from_forked_origin.vs["swhid"][e.target]].append(
            inventory_from_forked_origin.vs["swhid"][e.source]
        )

    def find_recent_references(
        target: ExtendedSWHID, limit: int
    ) -> List[ExtendedSWHID]:
        return references[target]

    # Recent API, see:
    # https://gitlab.softwareheritage.org/swh/devel/swh-storage/-/merge_requests/1042
    mocker.patch.object(
        sample_populated_storage,
        "object_find_recent_references",
        create=True,
        side_effect=find_recent_references,
    )
    return sample_populated_storage


@pytest.mark.parametrize(
    "graph_client_fixture, storage_fixture",
    [
        (
            "graph_client_with_both_origins",
            "storage_with_no_new_references_since_export",
        ),
        (
            "graph_client_with_only_initial_origin",
            "storage_with_references_from_forked_origin",
        ),
        (
            "graph_client_with_both_origins",
            "storage_with_references_from_forked_origin",
        ),
    ],
)
def test_mark_removable_on_initial_origin(
    request, storage_fixture, graph_client_fixture, inventory_from_initial_origin
):
    storage = request.getfixturevalue(storage_fixture)
    graph_client = request.getfixturevalue(graph_client_fixture)
    subgraph = mark_removable(storage, graph_client, inventory_from_initial_origin)
    # We are trying to remove the initial origin. As everything is used by the
    # fork, the only things that can be removed in the end are the corresponding
    # origin and snapshot.
    assert {str(swhid) for swhid in subgraph.removable_swhids()} == {
        "swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054",
        "swh:1:snp:0000000000000000000000000000000000000020",
    }


@pytest.mark.parametrize(
    "graph_client_fixture, storage_fixture",
    [
        (
            "graph_client_with_both_origins",
            "storage_with_no_new_references_since_export",
        ),
        (
            "graph_client_with_only_initial_origin",
            "storage_with_references_from_forked_origin",
        ),
        (
            "graph_client_with_both_origins",
            "storage_with_references_from_forked_origin",
        ),
        (
            "graph_client_with_submodule",
            # This one is not ideal as it does not contain our origin with
            # submodule. This should not matter for testing that we have
            # the right behavior in presence of submodules, though.
            "storage_with_no_new_references_since_export",
        ),
    ],
)
def test_mark_removable_on_forked_origin(
    request, storage_fixture, graph_client_fixture, inventory_from_forked_origin
):
    storage = request.getfixturevalue(storage_fixture)
    graph_client = request.getfixturevalue(graph_client_fixture)
    subgraph = mark_removable(storage, graph_client, inventory_from_forked_origin)
    # We are trying to remove the forked origin. Therefore objects coming from the
    # initial origin are unremovable. So we are left removing all objects specific
    # to the development that happened in the forked origin.
    assert {str(swhid) for swhid in subgraph.removable_swhids()} == {
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:rel:0000000000000000000000000000000000000021",
        "swh:1:rev:0000000000000000000000000000000000000018",
        "swh:1:rev:0000000000000000000000000000000000000013",
        "swh:1:dir:0000000000000000000000000000000000000017",
        "swh:1:dir:0000000000000000000000000000000000000016",
        "swh:1:dir:0000000000000000000000000000000000000012",
        "swh:1:cnt:0000000000000000000000000000000000000015",
        "swh:1:cnt:0000000000000000000000000000000000000014",
        "swh:1:cnt:0000000000000000000000000000000000000011",
    }



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import logging
from typing import Iterator

import pytest

import swh.graph.example_dataset as graph_dataset
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Revision,
    RevisionType,
    Snapshot,
    SnapshotBranch,
    TargetType,
    Timestamp,
    TimestampWithTimezone,
)
from swh.model.swhids import ExtendedSWHID

from ..inventory import InventorySubgraph, Lister
from .test_subgraph import empty_subgraph, sample_data_subgraph  # noqa: F401

logger = logging.getLogger(__name__)


def h(id: int, width=40) -> bytes:
    return bytes.fromhex(f"{id:0{width}}")


@pytest.fixture
def sample_data_inventory(sample_data_subgraph):  # noqa: F811
    return InventorySubgraph.copy(sample_data_subgraph)


def test_select_incomplete_returns_only_incomplete(sample_data_inventory):
    g = sample_data_inventory
    # Mark everything as complete
    g.vs["complete"] = True
    v1 = g.vs.find("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165")
    v2 = g.vs.find("swh:1:dir:0000000000000000000000000000000000000012")
    v1["complete"] = False
    v2["complete"] = False
    assert set(g.select_incomplete()) == {v1, v2}


def test_select_incomplete_returns_sorted_by_object_type(sample_data_inventory):
    g = sample_data_inventory
    # Mark everything as complete
    g.vs["complete"] = True
    v_ori = g.vs.find("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165")
    v_snp = g.vs.find("swh:1:snp:0000000000000000000000000000000000000020")
    v_rel = g.vs.find("swh:1:rel:0000000000000000000000000000000000000010")
    v_rev = g.vs.find("swh:1:rev:0000000000000000000000000000000000000009")
    v_dir = g.vs.find("swh:1:dir:0000000000000000000000000000000000000008")
    v_cnt = g.vs.find("swh:1:cnt:0000000000000000000000000000000000000007")
    for v in v_ori, v_snp, v_rel, v_rev, v_dir, v_cnt:
        v["complete"] = False
    assert g.select_incomplete() == [v_ori, v_snp, v_rel, v_rev, v_dir, v_cnt]


@pytest.fixture
def lister_with_populated_storage(
    sample_populated_storage, graph_client_with_both_origins
):
    return Lister(
        sample_populated_storage, graph_client_with_both_origins, InventorySubgraph()
    )


def fix_contents(contents: Iterator[Content]) -> Iterator[Content]:
    """Recreate more complete Content objects using the same SWHIDs as the ones given.

    The content objects provided by :py:module:`swh.graph.example_dataset` are not
    complete enough to be inserted in a ``swh.storage``, so we make up what’s missing
    here."""
    for content in contents:
        swhid_value = int.from_bytes(content.swhid().object_id, "big")
        yield Content.from_dict(
            {
                "sha1": bytes.fromhex(f"{swhid_value:040x}"),
                "sha1_git": bytes.fromhex(f"{swhid_value:040x}"),
                "sha256": bytes.fromhex(f"{swhid_value:064x}"),
                "blake2s256": bytes.fromhex(f"{swhid_value:064x}"),
                "data": b"",
                "length": 0,
                "ctime": datetime.datetime.now(tz=datetime.timezone.utc),
            }
        )


@pytest.fixture
def snapshot_20_with_multiple_branches_pointing_to_the_same_head():
    # No snapshot in the example dataset has multiple branches
    # or tags pointing to the same head. It’s pretty common in
    # the real world though, so let’s have a test with that.
    return Snapshot(
        id=h(20),
        branches={
            b"refs/heads/master": SnapshotBranch(
                target=h(9), target_type=TargetType.REVISION
            ),
            b"refs/heads/dev": SnapshotBranch(
                target=h(9), target_type=TargetType.REVISION
            ),
            b"refs/tags/v1.0": SnapshotBranch(
                target=h(10), target_type=TargetType.RELEASE
            ),
        },
    )


@pytest.fixture
def directory_6_with_multiple_entries_pointing_to_the_same_content():
    # No directories in the example dataset has multiple entries
    # pointing to the same content. It can happen in the real world,
    # so let’s test that situation.
    return Directory(
        id=h(6),
        entries=(
            DirectoryEntry(
                name=b"README.md",
                perms=0o100644,
                type="file",
                target=h(4),
            ),
            DirectoryEntry(
                name=b"parser.c",
                perms=0o100644,
                type="file",
                target=h(5),
            ),
            DirectoryEntry(
                name=b"parser_backup.c",
                perms=0o100644,
                type="file",
                target=h(5),
            ),
        ),
    )


@pytest.fixture
def sample_populated_storage(
    swh_storage,
    snapshot_20_with_multiple_branches_pointing_to_the_same_head,
    directory_6_with_multiple_entries_pointing_to_the_same_content,
):
    result = swh_storage.content_add(fix_contents(graph_dataset.CONTENTS))
    assert result.get("content:add") == 6
    result = swh_storage.skipped_content_add(graph_dataset.SKIPPED_CONTENTS)
    assert result == {"skipped_content:add": 1}
    directories = list(graph_dataset.DIRECTORIES)
    directories[1] = directory_6_with_multiple_entries_pointing_to_the_same_content
    result = swh_storage.directory_add(directories)
    assert result == {"directory:add": 6}
    result = swh_storage.revision_add(graph_dataset.REVISIONS)
    assert result == {"revision:add": 4}
    # swh.graph.example_dataset contains a dangling release which would
    # prevent us from removing any revisions, directories or contents in our tests.
    # We need to skip it.
    result = swh_storage.release_add(
        [
            rel
            for rel in graph_dataset.RELEASES
            if str(rel.swhid()) != "swh:1:rel:0000000000000000000000000000000000000019"
        ]
    )
    assert result == {"release:add": 2}
    snapshot_22 = graph_dataset.SNAPSHOTS[1]
    result = swh_storage.snapshot_add(
        [snapshot_20_with_multiple_branches_pointing_to_the_same_head, snapshot_22]
    )
    assert result == {"snapshot:add": 2}
    result = swh_storage.origin_add(graph_dataset.ORIGINS)
    assert result == {"origin:add": 2}
    swh_storage.origin_visit_add(graph_dataset.ORIGIN_VISITS)
    swh_storage.origin_visit_status_add(graph_dataset.ORIGIN_VISIT_STATUSES)
    return swh_storage


@pytest.fixture
def graph_client_with_only_initial_origin(naive_graph_client):
    from swh.graph.http_naive_client import NaiveClient

    initial_origin = str(graph_dataset.INITIAL_ORIGIN.swhid())
    return NaiveClient(
        nodes=list(naive_graph_client.visit_nodes(initial_origin)),
        edges=list(naive_graph_client.visit_edges(initial_origin)),
    )


@pytest.fixture
def graph_client_with_both_origins(naive_graph_client):
    from swh.graph.http_naive_client import NaiveClient

    # swh.graph.example_dataset contains a dangling release which would
    # prevent us from removing any revisions, directories or contents in our tests.
    # We skip it by reconstructing a graph from both origins
    initial_origin = str(graph_dataset.INITIAL_ORIGIN.swhid())
    forked_origin = str(graph_dataset.FORKED_ORIGIN.swhid())
    nodes = set(naive_graph_client.visit_nodes(initial_origin)) | set(
        naive_graph_client.visit_nodes(forked_origin)
    )
    edges = set(naive_graph_client.visit_edges(initial_origin)) | set(
        naive_graph_client.visit_edges(forked_origin)
    )
    return NaiveClient(nodes=nodes, edges=edges)


def assert_subgraph_is_full_from_forked_origin(subgraph):
    assert subgraph.to_list_dict(use_vids=False, sequence_constructor=set) == {
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165": {
            "swh:1:snp:0000000000000000000000000000000000000022"
        },
        "swh:1:snp:0000000000000000000000000000000000000022": {
            "swh:1:rel:0000000000000000000000000000000000000021",
            "swh:1:rel:0000000000000000000000000000000000000010",
            "swh:1:rev:0000000000000000000000000000000000000009",
        },
        "swh:1:rel:0000000000000000000000000000000000000021": {
            "swh:1:rev:0000000000000000000000000000000000000018"
        },
        "swh:1:rel:0000000000000000000000000000000000000010": {
            "swh:1:rev:0000000000000000000000000000000000000009"
        },
        "swh:1:rev:0000000000000000000000000000000000000018": {
            "swh:1:rev:0000000000000000000000000000000000000013",
            "swh:1:dir:0000000000000000000000000000000000000017",
        },
        "swh:1:rev:0000000000000000000000000000000000000013": {
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:dir:0000000000000000000000000000000000000012",
        },
        "swh:1:rev:0000000000000000000000000000000000000009": {
            "swh:1:rev:0000000000000000000000000000000000000003",
            "swh:1:dir:0000000000000000000000000000000000000008",
        },
        "swh:1:rev:0000000000000000000000000000000000000003": {
            "swh:1:dir:0000000000000000000000000000000000000002"
        },
        "swh:1:dir:0000000000000000000000000000000000000017": {
            "swh:1:dir:0000000000000000000000000000000000000016",
            "swh:1:cnt:0000000000000000000000000000000000000014",
        },
        "swh:1:dir:0000000000000000000000000000000000000016": {
            "swh:1:cnt:0000000000000000000000000000000000000015"
        },
        "swh:1:dir:0000000000000000000000000000000000000012": {
            "swh:1:dir:0000000000000000000000000000000000000008",
            "swh:1:cnt:0000000000000000000000000000000000000011",
        },
        "swh:1:dir:0000000000000000000000000000000000000008": {
            "swh:1:dir:0000000000000000000000000000000000000006",
            "swh:1:cnt:0000000000000000000000000000000000000007",
            "swh:1:cnt:0000000000000000000000000000000000000001",
        },
        "swh:1:dir:0000000000000000000000000000000000000006": {
            "swh:1:cnt:0000000000000000000000000000000000000005",
            "swh:1:cnt:0000000000000000000000000000000000000004",
        },
        "swh:1:dir:0000000000000000000000000000000000000002": {
            "swh:1:cnt:0000000000000000000000000000000000000001"
        },
    }


# Adding candidates using graph
# =============================


# Testing from an origin should be enough as it should travel the whole graph
def test_add_edges_traversing_graph_from_origin(lister_with_populated_storage):
    origin_swhid = ExtendedSWHID.from_string(
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"
    )
    lister_with_populated_storage.add_edges_traversing_graph(origin_swhid)
    subgraph = lister_with_populated_storage.subgraph
    assert_subgraph_is_full_from_forked_origin(subgraph)
    # But… the origin is still marked as incomplete as we want to see in swh.storage
    # if anything has been added since the swh.graph export
    assert len(subgraph.select_incomplete()) == 1


# Adding candidates using storage
# ===============================


def test_add_edges_using_storage_for_content(lister_with_populated_storage):
    content_swhid = ExtendedSWHID.from_string(
        "swh:1:cnt:0000000000000000000000000000000000000001"
    )
    lister_with_populated_storage.add_edges_using_storage(content_swhid)
    subgraph = lister_with_populated_storage.subgraph
    assert len(subgraph.vs) == 1
    assert (
        subgraph.vs.find("swh:1:cnt:0000000000000000000000000000000000000001")[
            "complete"
        ]
        is True
    )


def test_add_edges_using_storage_for_directory(lister_with_populated_storage):
    directory_swhid = ExtendedSWHID.from_string(
        "swh:1:dir:0000000000000000000000000000000000000008"
    )
    lister_with_populated_storage.add_edges_using_storage(directory_swhid)
    subgraph = lister_with_populated_storage.subgraph
    assert subgraph.to_list_dict(use_vids=False, sequence_constructor=set) == {
        "swh:1:dir:0000000000000000000000000000000000000008": {
            "swh:1:cnt:0000000000000000000000000000000000000001",
            "swh:1:cnt:0000000000000000000000000000000000000007",
            "swh:1:dir:0000000000000000000000000000000000000006",
        }
    }
    assert [v["name"] for v in subgraph.select_incomplete()] == [
        "swh:1:dir:0000000000000000000000000000000000000006"
    ]


def test_add_edges_using_storage_for_revision(lister_with_populated_storage):
    revision_swhid = ExtendedSWHID.from_string(
        "swh:1:rev:0000000000000000000000000000000000000013"
    )
    lister_with_populated_storage.add_edges_using_storage(revision_swhid)
    subgraph = lister_with_populated_storage.subgraph
    assert subgraph.to_list_dict(use_vids=False, sequence_constructor=set) == {
        "swh:1:rev:0000000000000000000000000000000000000013": {
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:dir:0000000000000000000000000000000000000012",
        },
        "swh:1:rev:0000000000000000000000000000000000000009": {
            "swh:1:rev:0000000000000000000000000000000000000003",
            "swh:1:dir:0000000000000000000000000000000000000008",
        },
        "swh:1:rev:0000000000000000000000000000000000000003": {
            "swh:1:dir:0000000000000000000000000000000000000002"
        },
    }
    assert {v["name"] for v in subgraph.select_incomplete()} == {
        "swh:1:dir:0000000000000000000000000000000000000012",
        "swh:1:dir:0000000000000000000000000000000000000008",
        "swh:1:dir:0000000000000000000000000000000000000002",
    }


def test_add_edges_using_storage_for_revisions_with_common_parents(
    lister_with_populated_storage,
):
    newer_revision_swhid = ExtendedSWHID.from_string(
        "swh:1:rev:0000000000000000000000000000000000000018"
    )
    older_revision_swhid = ExtendedSWHID.from_string(
        "swh:1:rev:0000000000000000000000000000000000000009"
    )
    lister_with_populated_storage.add_edges_using_storage(newer_revision_swhid)
    lister_with_populated_storage.add_edges_using_storage(older_revision_swhid)
    subgraph = lister_with_populated_storage.subgraph
    assert {v["name"] for v in subgraph.select_incomplete()} == {
        "swh:1:dir:0000000000000000000000000000000000000017",
        "swh:1:dir:0000000000000000000000000000000000000012",
        "swh:1:dir:0000000000000000000000000000000000000008",
        "swh:1:dir:0000000000000000000000000000000000000002",
    }


@pytest.fixture
def revision_pointing_multiple_times_to_the_same_parent():
    return Revision(
        id=h(42),
        message=b"Weird revision",
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111177770,
                microseconds=0,
            ),
            offset_bytes=b"+0000",
        ),
        committer=graph_dataset.PERSONS[0],
        author=graph_dataset.PERSONS[2],
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111177770,
                microseconds=0,
            ),
            offset_bytes=b"+0000",
        ),
        type=RevisionType.GIT,
        directory=h(2),
        synthetic=False,
        metadata=None,
        parents=(
            h(3),
            h(3),
        ),
    )


def test_add_edges_using_storage_for_revisions_pointing_multiple_times_to_the_same_parent(
    sample_populated_storage,
    graph_client_with_both_origins,
    revision_pointing_multiple_times_to_the_same_parent,
):
    storage = sample_populated_storage
    storage.revision_add([revision_pointing_multiple_times_to_the_same_parent])
    lister = Lister(
        sample_populated_storage, graph_client_with_both_origins, InventorySubgraph()
    )
    revision_swhid = ExtendedSWHID.from_string(
        "swh:1:rev:0000000000000000000000000000000000000042"
    )
    lister.add_edges_using_storage(revision_swhid)
    subgraph = lister.subgraph
    assert subgraph.to_list_dict(use_vids=False, sequence_constructor=set) == {
        "swh:1:rev:0000000000000000000000000000000000000042": {
            "swh:1:rev:0000000000000000000000000000000000000003",
            "swh:1:dir:0000000000000000000000000000000000000002",
        },
        "swh:1:rev:0000000000000000000000000000000000000003": {
            "swh:1:dir:0000000000000000000000000000000000000002"
        },
    }


def test_add_edges_using_storage_for_release(lister_with_populated_storage):
    release_swhid = ExtendedSWHID.from_string(
        "swh:1:rel:0000000000000000000000000000000000000010"
    )
    lister_with_populated_storage.add_edges_using_storage(release_swhid)
    subgraph = lister_with_populated_storage.subgraph
    assert subgraph.to_list_dict(use_vids=False, sequence_constructor=set) == {
        "swh:1:rel:0000000000000000000000000000000000000010": {
            "swh:1:rev:0000000000000000000000000000000000000009"
        },
    }
    assert {v["name"] for v in subgraph.select_incomplete()} == {
        "swh:1:rev:0000000000000000000000000000000000000009"
    }


def test_add_edges_using_storage_for_snapshot(lister_with_populated_storage):
    snapshot_swhid = ExtendedSWHID.from_string(
        "swh:1:snp:0000000000000000000000000000000000000022"
    )
    lister_with_populated_storage.add_edges_using_storage(snapshot_swhid)
    subgraph = lister_with_populated_storage.subgraph
    assert subgraph.to_list_dict(use_vids=False, sequence_constructor=set) == {
        "swh:1:snp:0000000000000000000000000000000000000022": {
            "swh:1:rel:0000000000000000000000000000000000000021",
            "swh:1:rel:0000000000000000000000000000000000000010",
            "swh:1:rev:0000000000000000000000000000000000000009",
        },
    }
    assert {v["name"] for v in subgraph.select_incomplete()} == {
        "swh:1:rel:0000000000000000000000000000000000000021",
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000009",
    }


def test_add_edges_using_storage_for_origin(lister_with_populated_storage):
    origin_swhid = ExtendedSWHID.from_string(
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"
    )
    lister_with_populated_storage.add_edges_using_storage(origin_swhid)
    subgraph = lister_with_populated_storage.subgraph
    assert subgraph.to_list_dict(use_vids=False, sequence_constructor=set) == {
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165": {
            "swh:1:snp:0000000000000000000000000000000000000022"
        },
    }
    assert {v["name"] for v in subgraph.select_incomplete()} == {
        "swh:1:snp:0000000000000000000000000000000000000022"
    }


#
# Inventory candidates
# ====================


@pytest.fixture
def lister_with_full_graph(lister_with_populated_storage):
    return lister_with_populated_storage


@pytest.fixture
def lister_with_empty_graph(sample_populated_storage):
    from swh.graph.http_naive_client import NaiveClient

    graph_client = NaiveClient(nodes=[], edges=[])
    return Lister(sample_populated_storage, graph_client, InventorySubgraph())


@pytest.fixture
def lister_with_partial_graph(
    sample_populated_storage, graph_client_with_only_initial_origin
):
    return Lister(
        sample_populated_storage,
        graph_client_with_only_initial_origin,
        InventorySubgraph(),
    )


# Depending on our sources, we will need to iterate more or less
# to inventory all candidates. These tests ensure that we do not
# perform too many iterations: if the graph is full, we hardly
# need to query the storage. The less complete the graph is,
# the more iterations gets needed.
@pytest.mark.parametrize(
    "fixture, max_iterations",
    [
        ("lister_with_full_graph", 3),
        ("lister_with_partial_graph", 13),
        ("lister_with_empty_graph", 13),
    ],
)
def test_inventory_candidates(request, caplog, fixture, max_iterations):
    lister = request.getfixturevalue(fixture)
    with caplog.at_level(logging.DEBUG):
        lister.inventory_candidates(graph_dataset.FORKED_ORIGIN.swhid())
        log_lines = [
            record
            for record in caplog.records
            if record.funcName == "inventory_candidates"
        ]
        assert len(log_lines) <= max_iterations
        caplog.clear()
    assert_subgraph_is_full_from_forked_origin(lister.subgraph)


#
# Submodules handling
# ===================


@pytest.fixture
def origin_with_submodule():
    # swh:1:ori:73186715131824fa4381c6b5ca041c1c90207ef0
    return Origin(url="https://example.com/swh/using-submodule")


@pytest.fixture
def sample_populated_storage_using_submodule(
    sample_populated_storage, origin_with_submodule
):
    # dir 30 → rev 13
    directory_with_submodule = Directory(
        id=h(30),
        entries=(
            DirectoryEntry(
                name=b"submodule",
                perms=0o100644,
                type="rev",
                # We pick a revision from the forked origin so we can test
                # if this has influence on our ability to remove the forked origin
                target=h(13),
            ),
        ),
    )
    revision_date = TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=1682496691,
            microseconds=0,
        ),
        offset_bytes=b"+0200",
    )
    # rev 31 → dir 30
    revision = Revision(
        id=h(31),
        message=b"Initial commit with submodule",
        date=revision_date,
        committer=graph_dataset.PERSONS[0],
        author=graph_dataset.PERSONS[0],
        committer_date=revision_date,
        type=RevisionType.GIT,
        directory=h(30),
        synthetic=False,
        metadata=None,
        parents=(),
    )
    # snp 32 → rev 31
    snapshot = Snapshot(
        id=h(32),
        branches={
            b"refs/heads/master": SnapshotBranch(
                target=h(31), target_type=TargetType.REVISION
            ),
        },
    )
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    visit = OriginVisit(
        origin="https://example.com/swh/using-submodule",
        date=now,
        visit=1,
        type="git",
    )
    visit_status = OriginVisitStatus(
        origin="https://example.com/swh/using-submodule",
        date=now,
        visit=1,
        type="git",
        status="full",
        snapshot=h(32),
        metadata=None,
    )
    storage = sample_populated_storage
    storage.directory_add([directory_with_submodule])
    storage.revision_add([revision])
    storage.snapshot_add([snapshot])
    storage.origin_add([origin_with_submodule])
    storage.origin_visit_add([visit])
    storage.origin_visit_status_add([visit_status])
    return storage


@pytest.fixture
def graph_client_with_submodule(naive_graph_client, origin_with_submodule):
    from swh.graph.http_naive_client import NaiveClient

    initial_origin = str(graph_dataset.INITIAL_ORIGIN.swhid())
    forked_origin = str(graph_dataset.FORKED_ORIGIN.swhid())

    extra_nodes = {
        origin_with_submodule,
        "swh:1:snp:0000000000000000000000000000000000000032",
        "swh:1:rev:0000000000000000000000000000000000000031",
        "swh:1:rev:0000000000000000000000000000000000000013",
        "swh:1:dir:0000000000000000000000000000000000000030",
    }
    extra_edges = {
        (origin_with_submodule, "swh:1:snp:0000000000000000000000000000000000000032"),
        (
            "swh:1:snp:0000000000000000000000000000000000000032",
            "swh:1:rev:0000000000000000000000000000000000000031",
        ),
        (
            "swh:1:rev:0000000000000000000000000000000000000031",
            "swh:1:dir:0000000000000000000000000000000000000030",
        ),
        (
            "swh:1:dir:0000000000000000000000000000000000000030",
            "swh:1:rev:0000000000000000000000000000000000000013",
        ),
    }
    nodes = (
        set(naive_graph_client.visit_nodes(initial_origin))
        | set(naive_graph_client.visit_nodes(forked_origin))
        | extra_nodes
    )
    edges = (
        set(naive_graph_client.visit_edges(initial_origin))
        | set(naive_graph_client.visit_edges(forked_origin))
        | extra_edges
    )
    return NaiveClient(nodes=nodes, edges=edges)


@pytest.mark.parametrize(
    "graph_client_fixture, storage_fixture",
    [
        (
            "graph_client_with_both_origins",
            "sample_populated_storage_using_submodule",
        ),
        (
            "graph_client_with_submodule",
            "sample_populated_storage_using_submodule",
        ),
    ],
)
def test_inventory_with_submodule_stops_at_directory(
    request, graph_client_fixture, storage_fixture, origin_with_submodule
):
    graph_client = request.getfixturevalue(graph_client_fixture)
    storage = request.getfixturevalue(storage_fixture)
    lister = Lister(storage, graph_client, InventorySubgraph())
    lister.inventory_candidates(origin_with_submodule.swhid())
    assert {str(swhid) for swhid in lister.subgraph.swhids()} == {
        "swh:1:ori:73186715131824fa4381c6b5ca041c1c90207ef0",
        "swh:1:snp:0000000000000000000000000000000000000032",
        "swh:1:rev:0000000000000000000000000000000000000031",
        "swh:1:dir:0000000000000000000000000000000000000030",
    }



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
This module implements the inventory stage of the
:ref:`removal algorithm <alter_removal_algorithm>`.
"""

from contextlib import suppress
import itertools
import logging
from typing import Any, Callable, Collection, Dict, Iterator, List

from igraph import Vertex

from swh.core.api.classes import stream_results_optional
from swh.graph.http_client import GraphArgumentException, RemoteGraphClient
from swh.model.model import Revision
from swh.model.swhids import ExtendedObjectType as ObjectType
from swh.model.swhids import ExtendedSWHID
from swh.storage.algos.origin import iter_origin_visit_statuses, iter_origin_visits
from swh.storage.algos.snapshot import snapshot_get_all_branches
from swh.storage.interface import StorageInterface

from .subgraph import Subgraph

logger = logging.getLogger(__name__)


class InventorySubgraph(Subgraph):
    """A subgraph holding an inventory of all candidates for removal

    When all references from a given node have been accounted for,
    the ``complete`` attribute is set to True.
    """

    default_vertex_attributes: Dict[str, Any] = {"complete": False}

    def __init__(self, *args, **kwargs):
        """See :py:class:`Subgraph`"""
        super().__init__(*args, **kwargs)
        self["name"] = "Inventory"
        self["root_swhids"] = []
        if "complete" not in self.vs.attributes():
            self.vs["complete"] = [False for _ in self.vs]

    def select_incomplete(self) -> List[Vertex]:
        """Return vertices known to be incomplete ordered by object type from
        origins to contents."""

        # We want to order incomplete vertices from origin to content in order to
        # increase the chance to complete as many vertices as possible using swh.graph
        # traversal which follows the same direction.
        return self.select_ordered(complete_eq=False)

    def dot_node_attributes(self, v: Vertex) -> List[str]:
        """Get a list of attributes in DOT format for the given vertex.

        On top of default attributes, use a bold font for root objects, and color the background
        if the vertex is known to be incomplete.

        :meta private:
        """
        attrs = super().dot_node_attributes(v)
        if v["swhid"] in self["root_swhids"]:
            attrs.append('fontname="times bold"')
        if not v["complete"]:
            attrs.append("color=gray")
        return attrs


class Lister:
    """A class encapsulating our inventory algorithm.

    :meta private:
    """

    def __init__(
        self,
        storage: StorageInterface,
        graph_client: RemoteGraphClient,
        subgraph: InventorySubgraph,
    ):
        self._subgraph = subgraph
        self._storage = storage
        self._graph_client = graph_client

    @property
    def subgraph(self):
        return self._subgraph

    def inventory_candidates(self, root: ExtendedSWHID) -> None:
        self._subgraph["root_swhids"].append(root)
        # Add our starting point as incomplete
        self._subgraph.add_swhid(root, complete=False)
        # Iterate until everything has been fetched
        logger.debug("inventory_candidates: added %s", root)
        for remaining in self._iter_inventory_candidates():
            logger.debug(
                "inventory_candidates: %4d SWIDS known, %4d needs to be looked up.",
                len(self._subgraph.vs),
                remaining,
            )

    def _fetch_candidates_using_graph(self, vertex) -> None:
        # We don’t except all SWHID to be present in swh.graph
        with suppress(GraphArgumentException):
            self.add_edges_traversing_graph(vertex["swhid"])

    def _fetch_candidates_using_storage(self, vertex) -> None:
        self.add_edges_using_storage(vertex["swhid"])

    def _iter_inventory_candidates(self) -> Iterator[int]:
        # We cycle from retrieving from swh.graph (fast but potentially incomplete),
        # and swh.storage (complete, but slow and one level at a time).
        # Note: SWHIDs discovered from swh.graph will not
        # require further fetching. Because objects are immutable,
        # nothing could have been added after the graph has been
        # exported. The sole exception is origin objects.
        # References from SWHIDs discovered from swh.storage
        # have to be looked up though. As there is a chance they
        # can be found them in swh.graph, we’ll do that at our
        # next iteration.
        fetchers = itertools.cycle(
            [self._fetch_candidates_using_graph, self._fetch_candidates_using_storage]
        )
        to_fetch = self._subgraph.select_incomplete()
        while to_fetch:
            fetch = next(fetchers)
            for vertex in to_fetch:
                # fetcher might complete more than a given vertex, so
                # this one might just have been completed since we got
                # the list of incomplete vertices
                if vertex["complete"]:
                    continue
                fetch(vertex)
            to_fetch = self._subgraph.select_incomplete()
            yield len(to_fetch)

    def add_edges_traversing_graph(self, start: ExtendedSWHID) -> None:
        # We want everything except dir→rev edges which represent submodules.
        # See the relevant comment in `_add_edges_using_storage_for_directory` below.
        edge_restriction = "ori:*,snp:*,rel:*,rev:*,dir:dir,dir:cnt"
        # XXX: We should be able to pass a SWHID object to visit_edges()
        for src, dst in self._graph_client.visit_edges(
            str(start), edges=edge_restriction
        ):
            # Origins are the only objects in the graph that are not identified
            # over their content. New destinations can be added over time with
            # new visits. Therefore, we need to always query storage for them.
            complete = not src.startswith("swh:1:ori:")
            v_src = self._subgraph.add_swhid(src, complete=complete)
            # Anything pointed by origins are known to be complete if they were
            # present at times the swh.graph export was made.
            v_dst = self._subgraph.add_swhid(dst, complete=True)
            self._subgraph.add_edge(v_src, v_dst)

    def add_edges_using_storage(self, source: ExtendedSWHID) -> None:
        _ADD_EDGES_USING_STORAGE_METHODS_PER_OBJECT_TYPE[source.object_type](
            self, source
        )

    def _add_edges_using_storage_for_content(self, source: ExtendedSWHID) -> None:
        _ = self._subgraph.add_swhid(source, complete=True)
        # Nothing is referenced by content objects so we have no edges to add

    def _add_edges_using_storage_for_directory(self, source: ExtendedSWHID) -> None:
        v_directory = self._subgraph.add_swhid(source, complete=True)
        entries = stream_results_optional(
            self._storage.directory_get_entries, source.object_id
        )
        if not entries:
            logger.warning("Directory %s not found", source)
            return
        for entry in entries:
            target = entry.swhid().to_extended()

            # References from directory to revision represents submodules. As of
            # April 2023, loaders will record the reference to a submodule, and
            # that’s it. They don’t handle loading the associated origin. If the
            # revision for a submodule is in the archive, it means that we the
            # origin has been retrieved separately. The policy is thus that to
            # remove the code in a submodule, one should remove the associated
            # origin. Long story short, revisions used as submodule (and their
            # references) are not considered as candidates for removal when
            # inventorying a directory.
            if target.object_type == ObjectType.REVISION:
                logger.info(
                    "Ignored submodule %s (%s), listed in directory %s",
                    entry.name,
                    entry.swhid(),
                    source,
                )
                continue

            # Content objects are our leafs. Therefore we already know all about
            # them as soon as we know they exist and we can consider them complete.
            if target.object_type == ObjectType.CONTENT:
                v_target = self._subgraph.add_swhid(target, complete=True)
            else:
                v_target = self._subgraph.add_swhid(target)
            # It is possible to get the same target multiple times
            # for a single directory. For example if it contains multiple
            # files with the same content.
            self._subgraph.add_edge(v_directory, v_target, skip_duplicates=True)

    def _add_edges_using_storage_for_revision(self, source: ExtendedSWHID) -> None:
        # We limit the search to not retrieve too much at once from storage.
        # We will have to retrieve the oldest parent revisions if needed in any cases.
        for rev_d in self._storage.revision_log([source.object_id], limit=100):
            revision = Revision.from_dict(rev_d)
            revision_swhid = revision.swhid().to_extended()
            # We might know about this revision already from a previous revision log.
            # In that case there is no need to go any further up the log.
            if self._subgraph.vs.select(name=str(revision_swhid), complete=True):
                break
            # We can say these don’t need to be fetched as we know that
            # we are getting everything about this revision here:
            # its directory and parents.
            v_revision = self._subgraph.add_swhid(revision_swhid, complete=True)
            v_directory = self._subgraph.add_swhid(revision.directory_swhid())
            self._subgraph.add_edge(v_revision, v_directory)
            # We create a set here because some revisions point to the same
            # parent multiple times.
            for parent_swhid in set(revision.parent_swhids()):
                v_parent = self._subgraph.add_swhid(parent_swhid)
                self._subgraph.add_edge(v_revision, v_parent)

    def _add_edges_using_storage_for_release(self, source: ExtendedSWHID) -> None:
        v_release = self._subgraph.add_swhid(source, complete=True)
        [release] = self._storage.release_get([source.object_id])
        if not release:
            logger.warning("Release %s not found", source)
            return
        v_target = self._subgraph.add_swhid(release.target_swhid())
        self._subgraph.add_edge(v_release, v_target)

    def _add_edges_using_storage_for_snapshot(self, source: ExtendedSWHID) -> None:
        v_snapshot = self._subgraph.add_swhid(source, complete=True)
        snapshot = snapshot_get_all_branches(self._storage, source.object_id)
        if not snapshot:
            logger.warning("Snapshot %s not found", source)
            return
        # We need to deduplicate targets as multiple branches can point to the
        # same head and we only need one edge for each target.
        # TODO: Better document aliases in swh-model. I had to look at mercurial
        #       loader to understand what they were useful for.
        # We also skip aliases here as they are referring to a branch name
        # inside the snapshot. As we inventory the reference for every branches
        # the branch pointed by an alias will always be dealt with.
        target_swhids = {
            branch.swhid()
            for branch in snapshot.branches.values()
            if branch and branch.target_type.value != "alias"
        }
        for swhid in target_swhids:
            v_branch = self._subgraph.add_swhid(swhid)
            self._subgraph.add_edge(v_snapshot, v_branch)

    def _add_edges_using_storage_for_origin(self, source: ExtendedSWHID) -> None:
        v_source = self._subgraph.add_swhid(source, complete=True)
        [origin_d] = self._storage.origin_get_by_sha1([source.object_id])
        if not origin_d:
            logger.warning("origin %s not found in storage", source)
            return
        for visit in iter_origin_visits(self._storage, origin_d["url"]):
            assert visit.visit is not None  # make mypy happy
            for status in iter_origin_visit_statuses(
                self._storage, visit.origin, visit.visit
            ):
                snapshot_swhid = status.snapshot_swhid()
                if snapshot_swhid:
                    v_snapshot = self._subgraph.add_swhid(snapshot_swhid)
                    self._subgraph.add_edge(v_source, v_snapshot, skip_duplicates=True)


_ADD_EDGES_USING_STORAGE_METHODS_PER_OBJECT_TYPE: Dict[
    ObjectType, Callable[[Lister, ExtendedSWHID], None]
] = {
    ObjectType.CONTENT: Lister._add_edges_using_storage_for_content,
    ObjectType.DIRECTORY: Lister._add_edges_using_storage_for_directory,
    ObjectType.REVISION: Lister._add_edges_using_storage_for_revision,
    ObjectType.RELEASE: Lister._add_edges_using_storage_for_release,
    ObjectType.SNAPSHOT: Lister._add_edges_using_storage_for_snapshot,
    ObjectType.ORIGIN: Lister._add_edges_using_storage_for_origin,
}


def make_inventory(
    storage, graph_client, swhids: Collection[ExtendedSWHID]
) -> InventorySubgraph:
    """Inventory candidates for removal from the given set of SWHID.

    By querying the given `storage` and `graph_client`, create a subgraph
    with all objects belonging to the given set of SWHIDs (typically origins).
    The result should then used to verify which candidate can safely be
    removed.
    """
    subgraph = InventorySubgraph()
    lister = Lister(storage, graph_client, subgraph)
    for swhid in swhids:
        lister.inventory_candidates(swhid)
    return lister.subgraph



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from typing import Any, Dict, List, TextIO

from igraph import Graph, Vertex

from swh.model.swhids import CoreSWHID
from swh.model.swhids import ExtendedObjectType as ObjectType
from swh.model.swhids import ExtendedSWHID

logger = logging.getLogger(__name__)


# Only used when generating graphviz output
def _batched(it, n):
    from itertools import islice

    # TODO: Migrate to the the assignment syntax once we require Python 3.8+
    # while batch := tuple(islice(it, n)):
    #     yield batch
    while True:
        batch = tuple(islice(it, n))
        if not batch:
            return
        yield batch


class Subgraph(Graph):
    """A class to hold a small subset of the Software Heritage graph

    Each vertex corresponds to a single SWHID and vice versa. The graph is
    directed and each edge represents a reference from an object in Software
    Heritage storage to another.

    This is backed by the `igraph <https://igraph.org/>`_ library for
    convenience.

    This class is intended to be subclassed to implement the more specific
    behaviors needed for the different stage of the :ref:`object removal
    algorithm <alter_removal_algorithm>`.
    """

    def __init__(self, *args, **kwargs):
        """See also `igraph.Graph constructor
        <https://igraph.org/python/api/latest/igraph.Graph.html#__init__>`_"""
        super().__init__(*args, directed=True, **kwargs)
        if "name" not in self.vs.attributes():
            self.vs["name"] = []
        if "swhid" not in self.vs.attributes():
            self.vs["swhid"] = []

    @classmethod
    def copy(cls, subgraph):
        """Create a new instance by copying vertices, edges and respective
        attributes from the given subgraph."""
        g = subgraph
        return cls(
            n=len(g.vs),
            edges=(e.tuple for e in g.es),
            graph_attrs={attr: g[attr] for attr in g.attributes()},
            vertex_attrs={attr: g.vs[attr] for attr in g.vs.attributes()},
            edge_attrs={attr: g.es[attr] for attr in g.es.attributes()},
        )

    # Used by select_ordered() below
    _OBJECT_TYPE_ORDER = {
        t: v
        for v, t in enumerate(
            [
                ObjectType.ORIGIN,
                ObjectType.SNAPSHOT,
                ObjectType.RELEASE,
                ObjectType.REVISION,
                ObjectType.DIRECTORY,
                ObjectType.CONTENT,
            ]
        )
    }

    default_vertex_attributes: Dict[str, Any] = {}
    """Vertex will get the following attributes on creation
    unless specified otherwise."""

    def add_vertex(self, name: str, **kwargs) -> Vertex:
        """Add or update a vertex.

        A vertex with the given `name` will be created if it does not exist already.

        Attributes for the vertex will be set to the one given as keyword arguments.

        Returns:
            a Vertex object corresponding to the added or updated vertex
        """
        try:
            # The name attribute is indexed. See:
            # https://lists.nongnu.org/archive/html/igraph-help/2014-05/msg00069.html
            v = self.vs.find(name)
            v.update_attributes(**kwargs)
            return v
        except ValueError:
            for k, v in type(self).default_vertex_attributes.items():
                kwargs[k] = kwargs.get(k, v)
            return super().add_vertex(name, source=set(), **kwargs)

    def add_swhid(self, object_or_swhid, **kwargs) -> Vertex:
        """Add or update a vertex for the given SWHID or object.

        This is a convenience method to add vertex from either :py:class:`CoreSWHID`,
        :py:class:`ExtendedSWHID`, or any objects implementing a ``swhid()`` method
        returning one of those.
        """
        if hasattr(object_or_swhid, "swhid") and callable(object_or_swhid.swhid):
            swhid = object_or_swhid.swhid()
        else:
            swhid = object_or_swhid
        if type(swhid) is CoreSWHID:
            swhid = swhid.to_extended()
        elif type(swhid) is str:
            swhid = ExtendedSWHID.from_string(swhid)
        return self.add_vertex(name=str(swhid), swhid=swhid, **kwargs)

    def swhids(self) -> List[ExtendedSWHID]:
        """Returns all SWHID in this subgraph"""
        return self.vs["swhid"]

    _DEBUG_EDGE_INSERTION = False

    def add_edge(self, src: Vertex, dst: Vertex, skip_duplicates=False, **kwargs):
        """Add an edge with the given attributes.

        When trying to add an edge that already exists:
        - if `skip_duplicates` is set to True, nothing will be done,
        - otherwise (the default), an exception will be raised.

        Raises:
            ValueError if the given edge already exists and `skip_duplicates` is False
        """
        if self.get_eid(src, dst, error=False) != -1:
            if skip_duplicates:
                return

            import inspect

            current_frame = inspect.currentframe()
            if current_frame and current_frame.f_back:
                caller = inspect.getframeinfo(current_frame.f_back)[2]
            else:
                caller = "<unknown>"
            raise ValueError(
                "Duplicate edge %s → %s, added from %s"
                % (src["name"], dst["name"], caller)
            )
        # This will also log edge insertion in the graph, making for a very
        # verbose output. It should only be useful if you get the
        # “duplicate edge” above.
        if False:
            import inspect

            logger.debug(
                "Inserting edge %s → %s, added from %s",
                src["name"],
                dst["name"],
                inspect.getframeinfo(inspect.currentframe().f_back)[2].strip(),
            )
        super().add_edge(src, dst, **kwargs)

    def select_ordered(self, *args, **kwargs) -> List[Vertex]:
        """Get vertices ordered by object type from origins to contents"""
        return sorted(
            self.vs.select(*args, **kwargs),
            key=lambda v: Subgraph._OBJECT_TYPE_ORDER[v["swhid"].object_type],
        )

    def _format_swhid_label(self, s):
        """Format SWHID taking only SHA1 and breaking in lines of 8 characters"""
        return "\\n".join(
            ["".join(batch) for batch in _batched(iter(s.split(":")[3]), 8)]
        )

    def dot_node_attributes(self, v: Vertex) -> List[str]:
        """Get a list of attributes in DOT format for the given vertex.

        The default implementation defines a label with a formatted SWHID.
        This method is called by :py:meth:`write_dot` and is meant to be
        subclassed to produce extra labels to highlight certain graph aspects.
        """
        return [f'label="{self._format_swhid_label(v["name"])}"']

    def write_dot(self, f: TextIO) -> None:
        """Write a representation of this subgraph in DOT format to `f`.

        The result can be processed using the `dot` command-line utility provided
        by the Graphviz package.
        """

        def write_objects(f, object_type):
            for v in self.vs.select(lambda v: v["swhid"].object_type == object_type):
                f.write(f'    {v.index} [{", ".join(self.dot_node_attributes(v))}];\n')
                for dst in v.successors():
                    # layout rev -> rev edges horizontally
                    if (
                        v["swhid"].object_type == ObjectType.REVISION
                        and dst["swhid"].object_type == ObjectType.REVISION
                    ):
                        edge_options = " [constraint=false]"
                    else:
                        edge_options = ""
                    f.write(f"    {v.index} -> {dst.index}{edge_options};\n")

        f.write(
            f'digraph "{self["name"] if "name" in self.attributes() else "Subgraph"}"'
            "{\n"
            "  ranksep=1; nodesep=0.5;\n"
            "  subgraph cnt {\n"
            "    node [style=filled, fillcolor=pink];\n"
        )
        write_objects(f, ObjectType.CONTENT)
        f.write(
            "  }\n"
            "  subgraph cluster_dir {\n"
            '    label="File contents";\n'
            "    node [shape=folder, style=filled, fillcolor=lightblue];\n"
        )
        write_objects(f, ObjectType.DIRECTORY)
        f.write(
            "  }\n"
            "  subgraph cluster_rev {\n"
            '    label="Revisions";\n'
            "    node [shape=diamond, style=filled, fillcolor=orchid];\n"
        )
        write_objects(f, ObjectType.REVISION)
        f.write(
            "  }\n"
            "  subgraph cluster_rel {\n"
            '    label="Releases";\n'
            "    node [shape=octagon, style=filled, fillcolor=sandybrown];\n"
        )
        write_objects(f, ObjectType.RELEASE)
        f.write(
            "  }\n"
            "  subgraph cluster_snp {\n"
            '    label="Snapshot";\n'
            "    node [shape=doubleoctagon, style=filled, fillcolor=aqua];\n"
        )
        write_objects(f, ObjectType.SNAPSHOT)
        f.write(
            "  }\n"
            "  subgraph cluster_ori {\n"
            '    label="Origins";\n'
            "    node [shape=egg, style=filled, fillcolor=khaki];\n"
        )
        write_objects(f, ObjectType.ORIGIN)
        f.write("  }\n}\n")
